/*
 * File : 
 * 	client/aioli.c
 * 	
 * 	aIOLi is an efficient I/O library for parallel access to remote storages in SMP clusters. 
 * 	Thanks to the kernel features, our framework provides parallel I/O algorithms without 
 * 	inter-processes synchronisation mechanisms as well as a simple interface based on 
 * 	the classic UNIX system calls (creat/open/read/write/close).
 *
 * 	This file is included in the first version of our prototype aIOLi.
 * 	To know which version is currently stable please visit http://aioli.imag.fr
 * 	
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author : 
 * 	Adrien Lebre, adrien.lebre@imag.fr
 *
 * Licence : 
 * 	GPLv2	
 * 	This program is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU General Public License
 * 	as published by the Free Software Foundation; either version 2
 * 	of the License, or (at your option) any later version.
 * 	
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program; if not, write to the Free Software
 * 	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Contributors :
 * 	BULL Open Source Division - INRIA France
 * 	
 * Created :
 * 	April 2004, the 1st
 *
 * Last Modification(s) : 
 * 	April 2005
 * 	
 * Compilation :
 * 	 see Makefile
 * 	
 * Comments : 
 * 	Implement the overloaded functions called in aioli.h
 * 	in order to exchange I/O requests between user process and aIOLi daemon.
 * 	
 * 	Actually, the read_aioli make only one copy : Indeed, a shared memory segment (with mmap) is create during
 * 	the initialisation phase. aIOLi uses this shared segment to put each related data from aiolid to client part.
 * 	The post_readrqst() have just to copy from shared memory segment to the real buffer provided by read_aioli call.
 * 	Note this approach is simple but not thread safe. Indeed, if two thread make a read, the shared memory segment is n't protected.
 *
 * Todo : 
 * 	grep TODO in file ;) ..... :D 
 *    check aiolid startup : fork a process / create aiolid / aioli client part wait for 1 sec (sleep(1)).
 *    Fix that by using SIGUSR1 sending by aiolid to aiolic when initialisation phase is over...
 * 	overloaded write() call.
 * 	overloaded dup() methods (corresponds to a simple copy of the relatd index in the iobx_tab ;) )
 * 	Please secure each shared memory segment. Now the mask "0666" is used. Find a way to limit the
 * 	access to the shm only by aioli daemon.
 * 	Find a cleany way to create/setup/close/delete IPC message queue, shared memory segment.
 * 	For example if the user forgets to close all files before it exits you have to clean
 * 	each memory things ;)
 *
 * 	Warnings, for asynchrone operations like iread or iwrite find a way to check replies.
 * 	Check if a solution exist without thread concept :p. I'm not sure but I don't think create a "user"
 * 	thread without the agreement of the user process will be a great idea :D.
 * 
 * Changelog :
 * 	look into the CVS log !
 *
*/

#include "aioli-shared.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <errno.h>

#include <unistd.h>
#include <getopt.h>

#include <sys/stat.h>
#include <fcntl.h>

#include <assert.h>

// IPC message queue and shared memory
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/shm.h>

// mmap things
#include <sys/types.h>
#include <sys/mman.h>

// Caught SIGINT and SIGQUIT
#include <signal.h>

#include <sys/time.h>
#include <time.h>

// for va_start() and derivated functions
#include <stdarg.h>

#define DIE(format,args...) do {		\
	char message[255];	\
	sprintf(message, "Error: ");		\
	sprintf(message, format "\n",## args);	\
	perror(message); \
	exit(1);				\
} while (0)

#define SHOWINFO 0

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// PROTOTYPES PART
//
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 

int creat_aioli(const char *pathname, mode_t mode);
int open_aioli (const char *pathname, int flags, ...);
int close_aioli(int fd);
off_t lseek_aioli(int fd, off_t offset, int whence);
size_t read_aioli(int fd, void *buf,size_t count);

int init_request_msgqueue(key_t key);
int post_request(int msg_queueid, long type, void *packet, size_t packet_size);
int init_reply_msgqueue(key_t key);
int get_reply(int msg_queueid, long type, void *packet, size_t packet_size);

int init_shm_buffer(const char* name,void **addr, unsigned long requested_size);
void init_aioli_clientpart(pid_t pid, key_t msg_queuekey, key_t shm_reply_key);
void clean_ipc(int signal);
void final_clean_ipc(void);
int post_initrqst(int msg_queueid,int rqstid, pid_t pid,key_t msgq_replykey,const char *shm_name);
int post_finalrqst(int msg_queueid, int rqstid, int pid_desc, pid_t pid);
int post_openrqst(int msg_queueid, int rqstid,  int desc_pid, const char *pathname, int flags, mode_t mode, int localfd);
int post_closerqst(int msg_queueid, int rqstid, int desc_pid, int localfd, int daemonfd);
int post_readrqst(int msg_queueid, int rqstid, int pid_desc, int localfd, int daemonfd, void *buf, size_t count);

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// PROCESS_IOB STRUCTURE MANAGEMENT PART
//
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 

// Global variables :
process_iob iobx_tab[MAX_FILEOPEN]; 

void init_iobxtab(){
	int i;
	for (i=0; i<MAX_FILEOPEN ; i++){
		if(i<3)
			iobx_tab[i].daemonfd = i; // stdin = 0 / stdout = 1 / stderr = 2
		else
			iobx_tab[i].daemonfd = -1;
  		
		iobx_tab[i].process_position = 0;
		iobx_tab[i].daemon_position = 0;
	}
}

// WARNINGS, each parameter should be tested before to be passed.
void add_file(int userfd, off_t process_position, int daemonfd, off_t daemon_position){
	iobx_tab[userfd].process_position=process_position;	
	iobx_tab[userfd].daemonfd=daemonfd;	
	iobx_tab[userfd].daemon_position=daemon_position;	
}

void remove_file(int userfd){
	iobx_tab[userfd].process_position=0;
	iobx_tab[userfd].daemonfd=-1;	
	iobx_tab[userfd].daemon_position = 0;
}

int find_freefiledesc(){
	int index;
	for (index=0;index<MAX_FILEOPEN;index++){
		if (iobx_tab[index].daemonfd <0)
			return index;
	}
	// TODO set errno to  MAX_FILEOPEN (equivalent to EMFILE)
	errno=EMFILE;	
	return -1;
}

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// MESSAGE QUEUE AND SHARED MEMORY PARTS
//
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 

// Global message queue variables :
int requestid = -1; //for each request posted requestid is incremented !

int msgq_requestid = -1;
key_t msgq_requestkey=MESSAGE_QUEUE_KEY;

int msgq_replyid = -1;	  
key_t msgq_replykey; 	//msgq_replykey = getpid();

// Global shared  memory variables : Please see aioli-shared.h

// Global miscellanous variables :
int pid_desc = -1 ; 	// the pid descriptor in aIOLi daemon

void final_clean_ipc(void){
	clean_ipc(-1);
}

int already_clean=0;
void clean_ipc(int signal){
	int result=-1;
	if(!already_clean){
#if SHOWINFO > 2	  
		fprintf(stderr,"clean IPC message queue(s)\n");
#endif
		already_clean=1;

		result=post_finalrqst(msgq_requestid,++requestid, pid_desc,getpid());
#if SHOWINFO
		if(result<0)
			fprintf(stderr,"Warnings, the processus was not properly unregistered\n");
		else
			fprintf(stderr,"Processus unregistered from aioli daemon\n");

		fprintf(stderr,"Clean IPC message queue(s)\n");
#endif
 		// Unlink shm_buffer
		if(shm_desc!=-1){
			if(shm_buffer!=NULL)
				munmap(shm_buffer,sizeof(reply_shm_obj)*MAX_SHMNB); 
			// munmap could be consider as useless since it will be at the end of the process but maybe a bit cleaner ;).
			close(shm_desc);
		}
	
		//Remove "reply message queue"
  		msgctl(msgq_replyid,IPC_RMID,NULL);

		if(signal!=-1)
			raise(signal);	 // redirect to processus
		else
			exit(0);
	}
	else{
#if SHOWINFO>2
		fprintf(sterr,"clean_ipc(): ipc semaphores and message queues have been already cleaned!\n");		
#endif	
	   exit(0);
	}
}

// Initialize the message queue to exchange request from user process (aioli client part) to aIOLi daemon.
// WARNINGS init_request_msgqueue return a good msg queue id or die so you don't have to check the returned value !!!
int init_request_msgqueue(key_t key){
	int msg_queueid;
	msg_queueid=msgget(key,0666);
	if(msg_queueid<0){
/*		if(errno==ENOENT){ // aIOLi seems to be stop. So we have to start it ;)
			system("\"aiolid > /tmp/aiolid_log 2>&1 &\""); // """ is required for redirect stdout and stderr since system is equal to sh -c. 
			msg_queueid=msgget(key, IPC_CREAT | 0666);
			if(msg_queueid<0){
					DIE("aioli.c: aioli daemon launched and can't find the message queue to send request");
			}
		}*/
	// A bit clean maybe with a fork ;) ...			
		int childpid;
		if(errno==ENOENT){ // aIOLi seems to be stop. So we have to start it ;)
			childpid=fork();
			if(childpid<0){
				fprintf(stderr,"aioli.c:  Can't fork aiolid, so die...\n");
				perror("");
				exit(0);
			}
			else if(childpid==0){ // Child part
			  	fprintf(stderr,"aioli.c:  gonna start aiolid ...\n");
				if(execlp("aiolid","aioli","-l","/tmp/aiolid_log",0)==-1){
			  		fprintf(stderr,"aioli.c:  Can't start aiolid, so die...\n");
					perror("");
				}
		  	  	exit(0);
			}
			else{ // aIOLi client part
				sleep(1); // TODO improve that by waiting on SIGUSR1...
				msg_queueid=msgget(key,IPC_CREAT | 0666);
				if(msg_queueid<0){
					fprintf(stderr,"aioli.c: aioli daemon launched and can't find the message queue to send request\n");
					perror("");
					exit(0);
				}
			}
		}
		else{
			fprintf(stderr,"aioli.c:  Something was wrong during message queue id recovery");
			perror("");
			exit(0);
		}
	}
	return msg_queueid;
}
// Post a specific request on the message queue definied by msg_queueid
int post_request(int msg_queueid, long type,void *packet, size_t packet_size){

	int result=-1;
	// First, initialize msg fields
	msg_request msg;
#if SHOWINFO
//	fprintf(stderr,"aioli.c: post_request() : Since ipc message queue required a type > 0 and INIT_REQ=0 we temporarily increment type\n");
#endif
	type++;	
	if(type<=0)
		DIE("Type of packet must be greater than 0, can't copy and send the request to the aioli daemon ! :(");
	else		  
		msg.type=type;
	if (packet_size<MAX_PACKETSZ)
		memcpy(msg.packet,packet,packet_size);
	else
		DIE("The size of packet is bigger than MAX_PACKETSZ, can't copy and send the request to the aioli daemon ! :(");
	
	result=msgsnd(msg_queueid, &msg, packet_size, 0);
	if(result<0){
		DIE("Can't post request to the message queue %d",msg_queueid);
	}
	return result;
}	

// Initialize the message queue to exchange reply from aIOLi daemon to user processus (aioli client part).
// WARNINGS init_reply_msgqueue return a good msg queue id or die so you don't have to check the returned value !!!
int init_reply_msgqueue(key_t key){
	int msg_queueid=-1;
	msg_queueid=msgget(key,IPC_CREAT | 0666);
	if(msg_queueid<0){
		DIE("aioli.c: init_reply_msgqueue():  Something was wrong during message queue id recovery errno=%d",errno);
	}
#if SHOWINFO	
	fprintf(stderr,"aioli.c:init_reply_msgqueue(): create an IPC message queue with key %d\n",key);
#endif	
	return msg_queueid;
}

int get_reply(int msg_queueid, long type, void *packet, size_t packet_size){
	msg_reply msg;
	if (msgrcv(msg_queueid, &msg, MAX_PACKETSZ, type, 0) >= 0){  // TODO check we could be uses some msgtype priorities. 
					  																//For example retrieves first read request, write and open and close so why not ;) (man msgrcv and check on msgtyp)
		memcpy(packet,msg.packet,packet_size);
		return 1;
	}
	else{
#if SHOWINFO
		fprintf(stderr,"aioli.c:get_reply():msgrcv errno=%d",errno);
		perror("");
#endif	
	return -1;
	}
}			  

// Creates an initializes a shared object. This object are going to be map to limit the copies between aioli client part and aIOLi daemon.
// Returns a shm descriptor and set the addr pointer to the shared memory, otherwise the processus DIE :p  
int init_shm_buffer(const char* name, void** addr,unsigned long requested_size){
	int fd=-1;

	/* Create a new memory object */
	fd = shm_open( name, O_RDWR, 0666 ); 
	if(fd==-1)
		DIE("aioli.c:init_shm_buffer(),Open failed (name=%s)\n",name);
			          

	/* Map the memory object */
	*addr=mmap( 0, requested_size,PROT_READ | PROT_WRITE,MAP_SHARED, fd, 0 );
	if(*addr==MAP_FAILED)
		DIE("aioli.c:init_shm_buffer():mmap failed\n");

#if SHOWINFO	
	fprintf(stdout,"aioli.c:init_shm_buffer(): Map addr gonna be returned is 0x%p\n",*addr);
#endif	
	return fd;
}

// Initialize the message queue and the shared memory reply segment
// (to exchange request from user process (aioli client part) to
//  aIOLi daemon and reciprocally for replies.
void init_aioli_clientpart(pid_t pid, key_t msgq_requestkey, key_t msgq_replykey){
	int index; 

	// TODO : 	it seems to be useless, terminaison signal have been caught below.
	// 			Moreover, it could generate a conflict beetween calling processus atexit() calls.
	 	
	if(atexit(final_clean_ipc)!=0)
		DIE("cannot set \"final_clean_ipc\" as an exit function\n");
	
	signal(SIGINT, &clean_ipc); // for CTRL+T // SIGSTOP and SIGKILL couldn't be caught.  
	// TODO analyse and check dependant with MPI layer and other libraries
	// Actually, if linked libraries already catch some signals, does aioli can ? ?? 
	signal(SIGQUIT, &clean_ipc);
	signal(SIGTERM, &clean_ipc);
	signal(SIGSEGV, &clean_ipc);
					
	// Create and setup the IPC message queue (return the message queue id).
	msgq_requestid=init_request_msgqueue(msgq_requestkey);
	
	// Create and attach the IPC shared memory segment (return a pointer to the process_reply variable)
	msgq_replykey=getpid();
	msgq_replyid=init_reply_msgqueue(msgq_replykey);

	// Get the semaphores to control access to the shm_buffer.
	for (index=0;index<MAX_SHMNB;index++){
		shm_mutex[index]=mutex_get(FIRST_MUTEX_KEY+index);
	} 
	
	// Create an map the shm_buffer (it enables to limit the copies between aioli client part and aIOLi daemon during `read` and `write` calls)	
  	shm_desc=init_shm_buffer(shm_name,&shm_buffer,sizeof(reply_shm_obj)*MAX_SHMNB);
	if(shm_desc<0)
		DIE("aioli.c:init_aioli_clientpart(): can't init_shm_buffer() (name=%s) :((\n",shm_name);
	shm_objs=shm_buffer;
	
	// Initialize different strcutures
	init_iobxtab();	
	
	// Send an init_request to aioli daemon.
	requestid=0;
	pid_desc=post_initrqst(msgq_requestid, getpid(), requestid, msgq_replykey,shm_name);
	if(pid_desc<0)
		DIE("aioli.c:init_aioli_clientpart():Bad pid descriptor returned by aioli daemon, so we can't use aioli client part :((");
}

int post_initrqst(int msg_queueid,pid_t pid, int rqstid, key_t msgq_replykey, const char* shm_name){
	init_request irqst;
	
	if (pid <0)
		DIE("pid must be a positive value");
	
	irqst.requestid=rqstid;
	irqst.pid=pid;
	irqst.msgq_replykey=msgq_replykey;

	strncpy(irqst.shm_name,shm_name,10);

	post_request(msg_queueid,INIT_REQ,&irqst,sizeof(init_request));
	
	generic_reply reply;
	get_reply(msgq_replyid,INIT_REPLY,&reply,sizeof(generic_reply));
	if(reply.result<0){
#if SHOWINFO			  
		fprintf(stderr,"aioli.c:post_initrqs():The request was bad :(\n"); 
#endif		
		errno=reply.remote_errno;
	}
	return reply.result;
	
}

int post_finalrqst(int msg_queueid, int rqstid, int pid_desc, pid_t pid){
	finalize_request frqst;

	if (pid <0)
		DIE("pid must be a positive value");

	frqst.requestid=rqstid;	
	frqst.pid_desc=pid_desc;	
	frqst.pid=pid;	

#if SHOWINFO
	fprintf(stderr,"aioli.c: post_finalrqst() from pid=%d (pid_desc=%d)\n",pid,pid_desc);
#endif

	post_request(msg_queueid,FINAL_REQ,&frqst,sizeof(finalize_request));
	
	generic_reply reply;
	get_reply(msgq_replyid,FINAL_REPLY,&reply,sizeof(generic_reply));
	if(reply.result<0){
#if SHOWINFO			  
		fprintf(stderr,"aioli.c:post_finalrqst():The request was bad or something was wrong durin get_reply()\n"); 
#endif		
		errno=reply.remote_errno;
	}
	return reply.result;
}

// WARNINGS : pathname must'nt be NULL !! (there is no test on it)
// Return remote fd >0, otherwise -1.
int post_openrqst(int msg_queueid,int rqstid, int pid_desc, const char *pathname, int flags, mode_t mode, int localfd){

	// First, initialize the different open_request fields : 
	open_request orqst;

	if(pid_desc<0)
		DIE("aioli.c:post_openrqst(): pid_desc must be a positive value\n");
	orqst.pid_desc=pid_desc;
	
	orqst.requestid=rqstid;

	orqst.userfd=localfd;	// TODO, if localfd equals -1, have to define the default behaviour. Now, just continue the post_openrqst....	
	
	if (strlen(pathname)<=MAX_FILENAMESZ-1) // just check overflow ;)
		strcpy(orqst.pathname,pathname);
	else
		DIE("size of filename is too long, so can't copy and pass to aioli deamon :(!\n");
	
	orqst.flags=flags;
	orqst.mode=mode;

	// Second : Post the request in the message queue
	post_request(msg_queueid, OPEN_REQ,&orqst, sizeof(open_request));	

	generic_reply reply;	
	get_reply(msgq_replyid,OPEN_REPLY,&reply,sizeof(generic_reply));
	if(reply.result<0)
		errno=reply.remote_errno;
	return reply.result;
	
}

int post_closerqst(int msg_queueid, int rqstid, int pid_desc, int localfd, int daemonfd){
	// First, initialize the different close_request fields : 
	close_request clrqst;
	
	clrqst.requestid=rqstid;
	clrqst.pid_desc=pid_desc;
	clrqst.userfd=localfd;	// TODO, if localfd equals -1, have to define the default behaviour. Now, just continue hte post_openrqst....	
	clrqst.daemonfd=daemonfd;

	// Second : Post the request in the message queue
	post_request(msg_queueid,CLOSE_REQ,&clrqst,sizeof(close_request));	
#if SHOWINFO
	fprintf(stderr,"Close request posted by processus %d (pid_desc=%d) for file (localfd=%d, daemonfd=%d)\n", getpid(),pid_desc,localfd,daemonfd);
#endif 	  	  
	generic_reply reply;
	get_reply(msgq_replyid,CLOSE_REPLY,&reply,sizeof(generic_reply));
	if(reply.result<0)
		errno=reply.remote_errno;
	return reply.result;

}

int post_readrqst(int msg_queueid, int rqstid, int pid_desc, int localfd, int daemonfd, void *buf, size_t count){

	// First Part : post read request: 
	read_request rrqst;
	
	rrqst.requestid=rqstid;
	rrqst.pid_desc=pid_desc;
	
	rrqst.userfd=localfd;
	rrqst.offset=iobx_tab[localfd].process_position;
	rrqst.length=count;
	
	rrqst.daemonfd=daemonfd;

	post_request(msg_queueid,READ_REQ,&rrqst,sizeof(read_request));	
	// Lseek Part, if required

	// Second part : Retrieve reply
	read_reply reply;	  
	get_reply(msgq_replyid,READ_REPLY,&reply,sizeof(read_reply));
	if(reply.result<0)
		errno=reply.remote_errno;
	else{
#if SHOWINFO
	fprintf(stderr,"aioli.c: post_readrqst() : Reply for pid_desc=%d: nb read byte(s)%d, shm_index=%d and seek start=%ld\n",rrqst.pid_desc,reply.result,reply.shm_index,reply.seek_start);
#endif			  
	   memcpy(buf,shm_objs[reply.shm_index].bytes+reply.seek_start,reply.result); //TODO skip this copy by a shared memory tips in order to make zero copy ;).
#if SHOWINFO
	fprintf(stderr,"aioli.c: post_readrqst() : memcpy for pid_desc=%d ok\n",rrqst.pid_desc);
#endif			  


		iobx_tab[localfd].process_position+=reply.result;  //TODO Handle global_position.
	}
//	fprintf(stderr,"aioli.c: post_readrqst() Reply for pid_desc=%d: nb read byte(s)%d, shm_index=%d and seek start=%ld, waiting for lock \n",rrqst.pid_desc,reply.result,reply.shm_index,reply.seek_start);
	mutex_lock(shm_mutex[reply.shm_index]);
//fprintf(stderr,"aioli.c: post_readrqst() :pid_desc=%d reply.shm_index=%d and link=%d\n",rrqst.pid_desc,reply.shm_index,shm_objs[reply.shm_index].link;

//	assert(shm_objs[reply.shm_index].link>=1);
	if(shm_objs[reply.shm_index].link<1)
			  abort();
//	
	--shm_objs[reply.shm_index].link;
	//x=msync(&(shm_objs[reply.shm_index].link),sizeof(int),MS_SYNC);
	//if(x!=0)
	mutex_unlock(shm_mutex[reply.shm_index]);
//	fprintf(stderr,"aioli.c: post_readrqst() pid_desc= %d : rqstiid=%d : link %d\n",rrqst.pid_desc,rqstid,x);
#if SHOWINFO
	fprintf(stderr,"aioli.c: post_readrqst() : unlock shm mutex ok too for pid_desc=%d \n",rrqst.pid_desc);
#endif			  
	return reply.result;
}

int required_aioli(int fd){
	return ((fd<MAX_FILEOPEN) && (iobx_tab[fd].daemonfd!=-1));
}
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// LIBC OVERLOAD PART
//
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 

// //// CREAT //// //
int creat_aioli(const char *pathname, mode_t mode){
	 return open_aioli (pathname, O_WRONLY|O_CREAT|O_TRUNC, mode);
}

// //// OPEN //// //
int open_aioli (const char *pathname, int flags, ...){

	char *absolute_path;
	int fd = -1;
	int result=-1;
	mode_t mode = 0;
	struct stat tempbuf; // Required to find the file type (socket/file/...)
	
#if SHOWINFO
	 	  fprintf(stderr,"aioli.c:open_aioli() called\n");
#endif 

	if (pathname == NULL){
		errno=EINVAL;
		return -1;
	}
	else if (stat(pathname,&tempbuf)<0){
		return -1;
	}
	
	// We only overload open for regular file ; other devices such as socket / specific character devices / ... are seimply redirected to the real open call 
	if(!S_ISREG(tempbuf.st_mode)){
		fd=open(pathname,flags,mode);
	}	
	else{
		if (flags & O_CREAT){
   		va_list arg;
			va_start(arg, flags); 
			mode = va_arg(arg, int);
			va_end(arg);
		}
	
		// First, check if the message queue is active (in another way, if library has been initialized ;))
		if(msgq_requestid<0){
			init_aioli_clientpart(getpid(),msgq_requestkey,msgq_replykey);
		}

		// Make local open to retrieve the file descirptor.
		fd=open(pathname,flags,mode);
		// We planned to skip this step by a shorter. Indeed, we just want to know the next free file descriptor.
		// However , for dup and other related calls, we couldn't assign whatever file descriptor, we will fix that during the next version.
		//fd=find_freefiledesc();
		
		if(fd>=0){ // open didn't fail so continue...
			// Second, have to send an open request to aIOLi daemon.
			absolute_path=realpath(pathname,NULL);	
			if(absolute_path==NULL){
#if SHOWINFO		
				DIE("aioli.c: couldn't resolve the absolute pathname for %s (abs_path=%s)",pathname,absolute_path);	
#endif
			}
			result=post_openrqst(msgq_requestid, ++requestid, pid_desc, absolute_path, flags, mode, fd);
			if (result<0){
#if SHOWINFO		
				DIE("aioli.c: aioli daemon couldn't open the file %s", absolute_path);	
#endif
			}
			else{ 
#if SHOWINFO
	 	  		fprintf(stderr,"aioli.c:open_aioli(): aiolid has opened file %s with daemon_file_desc =%d\n", absolute_path,result);
#endif 
	  	 	 	add_file(fd,0,result,0);  // TODO WARNINGS, actually the daemon_position is set brutally to 0. We have to implement
		  									  // a IPC message queue for the replies and return more thant only one value
			}
		}
	}
#if SHOWINFO
	fprintf(stderr,"aioli.c:open_aioli() gonna return %d\n", fd);
#endif 
	return fd;
}

// //// CLOSE //// //
int close_aioli(int fd){
#if SHOWINFO
	 	  fprintf(stderr,"aioli.c:close_aioli() called\n");
#endif 
	if(!required_aioli(fd)){
		return close(fd);
	}
	else{	
	   // TODO please, check if this is the last file you have to close :
		// 1./ detach and delete the different shm segment.
		// 2./ ....

#if SHOWINFO
		fprintf(stderr,"aioli.c: Inside close_aioli msgq_requestid=%d\n",msgq_requestid);
#endif
		// First, close in aioli daemon
		 int result=-1;
		 if (msgq_requestid >= 0){
		 	result=post_closerqst(msgq_requestid,++requestid, pid_desc, fd,iobx_tab[fd].daemonfd); 	// Return number of files still open in aioli daemon for this pid_desc.
			if(result>=0)
				remove_file(fd);
		 }
	
		 // Second, close in this processus.
		 return close(fd);
	}
}

// //// READ //// //
size_t read_aioli(int fd, void *buf,size_t count){
#if SHOWINFO
	 fprintf(stderr,"aioli.c : read_aioli() called\n ");
#endif	

	 if(!required_aioli(fd)){
		  return read(fd,buf,count);
	 }
	 else{
		  if (count == 0)
				return 0;
		  if (fd < 0){
				errno=EBADF;
				return -1;
		  }
		  if (buf == NULL){
				errno=EINVAL;
				return -1;
		  }
		  // Post read request and get the reply (set iobx_tab[fd].position to the right value ;) )
		  // So should have been called process_readrqst has in aIOLi but :p
		  return post_readrqst(msgq_requestid,++requestid,pid_desc,fd,iobx_tab[fd].daemonfd, buf, count);
	 }
}

off_t lseek_aioli(int fd, off_t offset, int whence){
#if SHOWINFO
	 fprintf(stderr,"aioli.c : lseek_aioli() called\n");
#endif	
	 if(!required_aioli(fd)){
		  return lseek(fd,offset,whence);
	 }
	 else{
		  if (fd < 0){
				errno=EBADF;
				return -1;
		  }
		  switch (whence)
		  {
				case SEEK_SET:
					 iobx_tab[fd].process_position=offset;
					 break;		  
				case SEEK_CUR:
					 iobx_tab[fd].process_position+=offset;
					 break;
				case SEEK_END:
					 DIE("aioli.c:lseek_aioli() whence SEEK_END is still not implemented :(, first 've to understand its purpose ;)");
					 break;  
				default:
					 errno=EINVAL;
#if SHOWINFO
					 fprintf(stderr,"\n");
#endif	
					 return -1;
		  }
#if SHOWINFO
		  fprintf(stderr,"set to offset %ld\n",iobx_tab[fd].process_position);
#endif	
		  return iobx_tab[fd].process_position;
	 }
}

