/*
 * File : 
 * 	daemon/aiolid.c
 * 	
 * 	aIOLi is an efficient I/O library for parallel access to remote storages in SMP clusters. 
 * 	Thanks to the kernel features, our framework provides parallel I/O algorithms without 
 * 	inter-processes synchronisation mechanisms as well as a simple interface based on 
 * 	the classic UNIX system calls (creat/open/read/write/close).
 *
 * 	This file is included in the first version of our prototype aIOLi.
 * 	To know which version is currently stable please visit http://aioli.imag.fr
 * 	
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author : 
 * 	Adrien Lebre, adrien.lebre@imag.fr
 *
 * Licence : 
 * 	GPLv2	
 * 	This program is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU General Public License
 * 	as published by the Free Software Foundation; either version 2
 * 	of the License, or (at your option) any later version.
 * 	
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program; if not, write to the Free Software
 * 	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Contributors :
 * 	BULL Open Source Division - INRIA France
 * 	
 * Created :
 * 	April 2004, the 1st
 * 
 * Last Modification(s) : 
 * 	April 2005
 * 	
 * Compilation :
 * 	 see Makefile
 *
 * Execution : 
 * 	Could be execute like a deamon but usually it's launched by the library optimizedio.
 * 	To change thread priorities, you have to launch with root UID.
 * 	
 * Comments : 
 * 	1 thread waits for init/finalize, open/creat/close/read/write requests and fills related linked list according to the IO server in charge and the request type
 * 	There are two list by file descriptor : readrqst_list and writerqst_list.
 * 	These lists are sorted by an ascending offset and length order
 * 	Besides, Four list are used to store init, finalise open and close requests.	
 * 	 
 *	 	1 (or more) threads act(s) as I/O daemons handling disk operation and
 *	   send replies to the clients when needed.
 *	   The I/O threads firstly try to 'flush' init, open, close and finalize lists and finally the read/write request.
 *
 * 	This server waits for requests sent by the aioli client partand
 *		generates answers for clients.
 *
 * Todo : 
 *  Grep TODO in file ;) ..... :D 
  * Handle write/dup requests.
 *  Included data stripping considerations
 *  and ..... 		
 *
 * Changelog :
 * 	look into the CVS log !
 *
*/

#ifndef _REENTRANT
# error "_REENTRANT must be defined for pthread support"
#endif

#define PROGNAME "aIOLi"


#include "aiolid.h"

#include <errno.h>
#include <getopt.h>

#include <stdlib.h>
#include <stdio.h>

//IPC message queue and shared memory
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/shm.h>

// mmap things
#include <sys/mman.h>

// Caught SIGINT and SIGQUIT to clean all ipc things ;)
#include <signal.h>

// getpid, getuid ...
#include <unistd.h>

// create, open, close, read ...
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

// strcpy
#include <string.h>

//
// sha1 - hash function
// #include <openssl/sha.h>
// Currently, not implemented, it will enable to retrieve more efficiently if a file is already open !
// See, add_file() function.
//

//  setpriority() 
#include <sys/time.h>
#include <sys/resource.h>


// readv and writev libc calls
#include <sys/uio.h>

// assert function
#include <assert.h>

// nanosleep function and gettimeofday
#include <time.h>
//#include <sys/time.h> Already defined above for setpriority


#include <endian.h>
#if (__BYTE_ORDER != __LITTLE_ENDIAN)
# error "Parts of this code depends on CPU endianness: it was only tested on Linux i386"
#endif


#define DIE(format,args...) do {		\
	char message[255];	\
	sprintf(message, "Error: ");		\
	sprintf(message, format "\n",## args);	\
	perror(message); \
	clean_ipc();				\
} while (0)


//#include "dbg.h"


#define isReadAggregatable(ptrbackward, ptrforward) \
			( (((read_request*)ptrbackward->packet)->offset+((read_request*)ptrbackward->packet)->length ) >= ((read_request*)(ptrforward->packet))->offset)

#define isWriteAggregatable(ptrbackward, ptrforward) \
			( (((write_request*)ptrbackward->packet)->offset+((write_request*)ptrbackward->packet)->length)==((write_request*)(ptrforward->packet))->offset)
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// PROTOTYPES PART
//
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 

// Global Variables
// notes : 'static' is used to limit the scope of this variables to this file .... so just static infact ;) 
static process_values process_tab[MAX_AGGREGATED_PROCESSES]; 		// Give informations for a specific process defined by pid_desc.
static pthread_mutex_t processes_mutex;									// Before to access to an unlocked process_tab cell, you have to hold this mutex.
																						// (Please, see process_initrqst() and add_process())

static rqst_list lists[NB_REQ_TYPE-2]; 								// Actually, we used 4 list (for init, finalyze, open and close request) ;
																					// read and write request are stored in each file_request.

static fdesc_info fdesc_tab[MAX_FILEOPEN];							// Give wich are working on a file defined by fd (or desc), absolute filename; global position and so on... .
static file_request fdesc_request[MAX_FILEOPEN];					// Stores read and write requests on each file.

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// SHARED MEMORY PARTS
//
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 

// Global Variables : Please see aioli-shared.h
static int logically_free_shm_obj;				// index to optimize the access to a free shm_obj (each time, it's incremented modulo MAX_SHMNB).
													// Indeed, logically, the free shm_obj is next ;) (circular buffer).


// Creates an initializes a shared object. This object are going to be map to limit the copies between optimizediolib and aioli daemon.
// Returns a shm_desc on the shared object and a pointer to the shared memory is set to shm_buffer , otherwise -1
int init_shm_buffer(const char *name, void **buffer, unsigned long requested_size){
	int fd=-1 ;	
	int result=-1;

	// Stupid but required manipulation to have 666 permission on shm_obj.
	// I don't know why but root seems to be not abble to create a 666 object directly. So Need to redefine umask just for the shm_object creation an after reset to backmode.
	mode_t backmode;
	backmode=umask(0);
	/* Create a new memory object */
	fd = shm_open( name, O_RDWR | O_CREAT,0666); 
	if(fd==-1){
		umask(backmode);
		perror("aiolid.c:init_shm_buffer(),Open failed\n");
		return -1;
	}
	umask(backmode);
	/* Set the memory object's size */
	result=ftruncate(fd,requested_size);
	if(result<0)
		DIE("aiolid.c:init_shm_buffer(): Can't set size %lu for the shm_buffer\n",requested_size);

	/* Map the memory object */
	*buffer=mmap( 0, requested_size,PROT_READ | PROT_WRITE,MAP_SHARED, fd, 0 );
	if(*buffer==MAP_FAILED){
		DIE("aiolid.c:init_shm_buffer():mmap failed\n");
	}
#if SHOWINFO > 1	
	fprintf(stdout,"aiolid.c:init_shm_buffer(): Map addr gonna be returned is 0x%p\n",shm_buffer);
#endif	
//	shm_unlink(shm_name); It should be done by the calling process
	return fd;
}


void init_shm_objs(void){

	int index; 

	shm_desc=-1;
	shm_buffer=NULL;	
	// First create shm_obj
	shm_desc=init_shm_buffer(shm_name, &shm_buffer,sizeof(reply_shm_obj)*MAX_SHMNB);
	if(shm_desc<0)
		DIE("aiolid.c:init_shm_objs() somethings wrong during init_shm_buffer call\n");
	
	shm_objs=shm_buffer;
	
	for (index=0;index<MAX_SHMNB;index++){
		shm_mutex[index]=mutex_init(FIRST_MUTEX_KEY+index);
		shm_objs[index].link=0;
	// shm_objs[index].bytes don't have to setup.
	} 
	logically_free_shm_obj=0;
}

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// MESSAGE QUEUE PARTS
//
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 

// Global variables
static int msgqid = -1;
static key_t msg_queuekey = MESSAGE_QUEUE_KEY;


// Initialize the message queue id to exchange request from user process (optimizedio library) to aioli daemon.
int init_request_msgqueue(key_t key){
	int msg_queueid;
	msg_queueid=msgget(key,IPC_CREAT | 0666);
	if(msg_queueid<0){
		DIE("aiolid.c: init_msgqueue(): Can't create an IPC message queue with key %d.",key);
	}
#if SHOWINFO > 1
	else
		fprintf(stderr, "aiolid.c: init_request_msgqueue(): global messagequeue id=%d well initialized\n", msg_queueid);
#endif

	return msg_queueid;
}

// Initialize the message queue id to exchange reply from aioli daemon to user process (optimizedio library).
// The difference with init_request_msgqueue() is the flag passed to msgget. Here the message queue have to be created by the user process and only by it ;).
int init_reply_msgqueue(key_t key){
	int msg_queueid;
	msg_queueid=msgget(key,0666);
	if(msg_queueid<0){
		DIE("aiolid.c: init_msgqueue(): Can't create an IPC message queue with key %d.",key);
	}
	return msg_queueid;
}

int post_genericreply(int msgqid,int rqstid, long type, int result, int error){
	
	generic_reply reply;

	reply.requestid=rqstid;
	reply.result=result;
	reply.remote_errno=error;
#if SHOWINFO > 1		
	fprintf(stderr,"aiolid.c: post_genericreply(): Try to post a \"generic\" reply (type=%ld, requestid=%d,result=%d) on message queue with id=%d\n",type, rqstid, result,msgqid);	
#endif 
	return post_reply(msgqid, type ,&reply,sizeof(generic_reply));
}

// Post a specific request on the message queue definied by msg_queueid
// WARNINGS : return 0 for right
int post_reply(int msg_queueid, long type,void *packet, size_t packet_size){
	int result=-1;

	// First, initialize msg fields
	msg_reply msg;
	msg.type=type;
	if (packet_size<MAX_PACKETSZ)
		memcpy(msg.packet,packet,packet_size);
	else{
#if SHOWINFO			  
	fprintf(stderr,"The size of packet is bigger than MAX_PACKETSZ, can't copy and send the request to the processus ! :(\n");
#endif	
		return result;
	}	
	result=msgsnd(msg_queueid, &msg,packet_size,0);
	if(result<0){
#if SHOWINFO			  
		fprintf(stderr,"Can't post request to the message queue %d\n",msg_queueid);
#endif 
		return result;
	}
	return result;
}

// Remove "request" ipc message queue
int already_clean=0;
void clean_ipc(void){
	int index;
	if(!already_clean){
#if SHOWINFO > 2	  
		fprintf(stderr,"clean IPC message queue(s)\n");
#endif
		already_clean=1;
		// semaphore-mutex on shm_obj have to be removed:
		for (index=0;index<MAX_SHMNB;index++){
			mutex_remove(shm_mutex[index]);
		} 
	
 	 	msgctl(msgqid,IPC_RMID,NULL); 

		// Unlink shm_buffer
		if(shm_desc!=-1){
			if(shm_buffer!=NULL)
				munmap(shm_buffer,sizeof(reply_shm_obj)*MAX_SHMNB); // Test it .... is it usefull or useless ... :)
			close(shm_desc);
			shm_unlink(shm_name);
		}
		system("rm -f /tmp/aiolid.pid");	
		fprintf(stderr,"aiolid has been stopped\n");		
		
	}
#if SHOWINFO>2
	else{
		fprintf(sterr,"clean_ipc(): ipc semaphores and message queues have been already cleaned!\n");		
	}
#endif	
	exit(0);
}

// WARNINGS, orqst parameters are not checked, overflow watchdog and kind of devil things are welcome ;)
void process_openrqst (open_request *orqst){
	int daemon_fd;
	
	if(process_tab[orqst->pid_desc].nb_file>=MAX_FILE_PERPROCESS){
#if SHOWINFO		  
		fprintf(stderr,"aiolid.c:process_openrqst():Error, can't add a new file to the process (pid=%d)\n",process_tab[orqst->pid_desc].pid);
#endif
		// TODO check How to handle new error value.
		errno=EN_FILEPERPROCESS;
		daemon_fd=-1; 
	}
	else
	{
		// Note : open_file_desc returns	daemon file descriptor

		daemon_fd=open_file_desc(orqst->pid_desc,orqst->pathname,orqst->flags,orqst->mode,orqst->userfd);
		if(daemon_fd<0){
#if SHOWINFO
			fprintf(stderr,"aiolid.c : process_openrqst() : Open request: filename:%s, user descriptor %d failed\n",orqst->pathname,orqst->userfd);
			perror("");
#endif
		}
		else{
#if SHOWINFO > 1
			fprintf(stderr,"aiolid.c : process_openrqst() : Open request: filename:%s, user descriptor %d, daemon descriptor%d\n",orqst->pathname,orqst->userfd,daemon_fd);
#endif
			process_tab[orqst->pid_desc].iob[orqst->userfd].process_position=0;
			process_tab[orqst->pid_desc].iob[orqst->userfd].daemonfd=daemon_fd;
			process_tab[orqst->pid_desc].nb_file++;
		}
	}
	if(post_genericreply(process_tab[orqst->pid_desc].msgq_replyid, orqst->requestid, OPEN_REPLY, daemon_fd, errno))
			  process_tab[orqst->pid_desc].last_ack_request=orqst->requestid;
	
}

void process_closerqst (close_request *clrqst){
	int result ;

	if(clrqst->pid_desc>MAX_AGGREGATED_PROCESSES){
#if SHOWINFO	
		fprintf(stderr,"aiolid.c: process_closerqst(), request arguments : bad pid descripteur %d\n",clrqst->pid_desc);
#endif
	 	errno= EBAD_PIDDESC;
		result = -1;
	}
	else if((clrqst->userfd>MAX_FILEOPEN)||(process_tab[clrqst->pid_desc].iob[clrqst->userfd].daemonfd != clrqst->daemonfd)){
#if SHOWINFO	
		fprintf(stderr,"aiolid.c: process_closerqst(), request arguments : Bad file descripteur %d for desc_pid=%d\n",clrqst->userfd,clrqst->pid_desc);
#endif
	 	errno= EBADF;
		result = -1;
	}
	else // args are right ;), so ....
	{
#if SHOWINFO > 1
		fprintf(stderr,"aiolid.c: process_closerqst(): Close request in progress : user descripteur %d, daemon descripteur %d\n",clrqst->userfd,clrqst->daemonfd);
#endif
		result=close_file_desc(clrqst->pid_desc,clrqst->userfd,clrqst->daemonfd);

		if (result >= 0){
		// The file has been correctly remove so update process_tab values
			process_tab[clrqst->pid_desc].iob[clrqst->userfd].daemonfd=-1;
			process_tab[clrqst->pid_desc].nb_file--;
			result = process_tab[clrqst->pid_desc].nb_file;
		//Note : We don't remove or assign to zero the different fields of the iob variable since the userfd is used has an index to this table.
		//So, next open will be overload this values ;). 
#if SHOWINFO > 2
		fprintf(stderr,"aiolid.c:process_closerqst(): processus %d (pid_desc=%d) have %d file(s) opened\n",process_tab[clrqst->pid_desc].pid,clrqst->pid_desc,process_tab[clrqst->pid_desc].nb_file);
#endif
		}
	}
	if(post_genericreply(process_tab[clrqst->pid_desc].msgq_replyid, clrqst->requestid, CLOSE_REPLY, result, errno))
		process_tab[clrqst->pid_desc].last_ack_request=clrqst->requestid;
}

void process_readrqst (read_request *rrqst){

	read_reply rreply;
	rreply.result = 1 ; 
	rreply.shm_index=-1;
	rreply.seek_start=0;

	// Find next free shm_obj 		// TODO WARNING Active wait. If all slots are used,this loops will never stop untill a sa new free slot... find somethings better
#if SHOWINFO > 2	
	fprintf(stderr,"aiolid.c:process_readrqst():next logically free shm_obj=%d\n",logically_free_shm_obj);
#endif
	mutex_lock(shm_mutex[logically_free_shm_obj]);
	while(shm_objs[logically_free_shm_obj].link>0){
		mutex_unlock(shm_mutex[logically_free_shm_obj]);
		logically_free_shm_obj=(logically_free_shm_obj+1)%MAX_SHMNB;
//DIE("aiolid.c:process_readrqsts(): search logically_free_shm_obj:  test %d wich contains link=%d\n",logically_free_shm_obj,shm_objs[logically_free_shm_obj].link);  
		mutex_lock(shm_mutex[logically_free_shm_obj]);
	}
#if SHOWINFO > 2	
	fprintf(stderr,"aiolid.c:process_readrqst(): selected shm_obj=%d\n",logically_free_shm_obj);
#endif
	rreply.shm_index=logically_free_shm_obj;
	logically_free_shm_obj=(logically_free_shm_obj+1)%MAX_SHMNB;
	assert(shm_objs[rreply.shm_index].link==0);
	shm_objs[rreply.shm_index].link=1; // Now, link > 0 so shm_objs[rreply.shm_index] is protected
	mutex_unlock(shm_mutex[rreply.shm_index]);

	// Lseek Part, if required
	if(pthread_mutex_lock(&fdesc_tab[rrqst->daemonfd].open_close_mutex)!=0){
		DIE("aiolid.c: pthread_mutex_lock Error");
	}
	if(rrqst->offset!=fdesc_tab[rrqst->daemonfd].global_position){
		rreply.result=lseek(rrqst->daemonfd,rrqst->offset,SEEK_SET);
		if(rreply.result<0){
			if(pthread_mutex_unlock(&fdesc_tab[rrqst->daemonfd].open_close_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
#if SHOWINFO	
			fprintf(stderr,"aiolid.c:process_readrqst():Can't set the fd=%d to the whished offset (offset=%ld)\n",rrqst->daemonfd,rrqst->offset);
			perror("");
#endif
		}
		else{
//			fdesc_tab[rrqst->daemonfd].global_position=rrqst->offset; // Useless since READ_RQST couldn't be parrallelized currently.
#if SHOWINFO > 2	
			fprintf(stderr,"aiolid.c:process_readrqst():fd=%d set to offset (offset=%ld)\n",rrqst->daemonfd,rrqst->offset);
#endif
		}
	}

	if(rreply.result!=-1){	
		// Read part
		rreply.result=read(rrqst->daemonfd,shm_objs[rreply.shm_index].bytes,rrqst->length);
		if (rreply.result <0){
			if(pthread_mutex_unlock(&fdesc_tab[rrqst->daemonfd].open_close_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_unlock Error");
			}
#if SHOWINFO	
			fprintf(stderr,"aiolid.c:process_readrqst(),read failed (from offset %ld in the file %s (fd=%d))\n",rrqst->offset,fdesc_tab[rrqst->daemonfd].pathname,rrqst->daemonfd);
			perror("");
#endif		
		}
		else{
#if SHOWINFO > 1	
			fprintf(stderr,"aiolid.c:process_readrqst(),read %d bytes from offset %ld in the file %s (fd=%d)\n",rrqst->length,rrqst->offset,fdesc_tab[rrqst->daemonfd].pathname,rrqst->daemonfd);
#endif
		fdesc_tab[rrqst->daemonfd].global_position+=rreply.result;
		if(pthread_mutex_unlock(&fdesc_tab[rrqst->daemonfd].open_close_mutex)!=0){
			DIE("aiolid.c: pthread_mutex_unlock Error");
		}
		// TODO cut and paste this line before real read call. This way enables to make several reads in a parallel way.
		// If read return <0 so go to the back (right) position ;) DEVIL or not :D - ACTUALLY, untill read request will be parralelized this is useless :(
		}
	}
	rreply.remote_errno=errno;
	rreply.requestid=rrqst->requestid;
	
	if(post_reply(process_tab[rrqst->pid_desc].msgq_replyid, READ_REPLY, &rreply, sizeof(read_reply))){
		process_tab[rrqst->pid_desc].last_ack_request=rrqst->requestid;
	}
}

void process_readrqsts (rqst_list* rrqsts){

   struct list_head *ptr_position; 			// Required for list_for_each_safe
	struct list_head *ptr_copy_position;	// Required for list_for_each_safe
	read_request* ptr_temp_rrqst;
	int nbrequest=0;
	int daemonfd;	
	off_t begin_offset,end_offset;
	size_t length;
	
	read_reply rreply;
	int global_result=1;
	rreply.result=-1;

	// First, the first request from rrqsts list provides the lower offset
	ptr_temp_rrqst=((read_request*)((rqst_cell*)list_entry(rrqsts->next,rqst_cell,list))->packet);
	begin_offset=ptr_temp_rrqst->offset;
	end_offset=begin_offset+ptr_temp_rrqst->length;
	daemonfd=ptr_temp_rrqst->daemonfd;
	
	list_for_each_safe(ptr_position,ptr_copy_position,rrqsts){
		ptr_temp_rrqst=((read_request*)((rqst_cell*)list_entry(ptr_position,rqst_cell,list))->packet);
#if SHOWINFO > 1
		fprintf(stderr,"aiolid.c : process_readrqsts() : offset=%ld, length=%lu \n",ptr_temp_rrqst->offset, ptr_temp_rrqst->length);
#endif		
		nbrequest++;
		// The highest offset could be provided from any requests
		if(end_offset < ptr_temp_rrqst->offset+ptr_temp_rrqst->length)
			end_offset= ptr_temp_rrqst->offset+ptr_temp_rrqst->length;
	}

	length= end_offset - begin_offset;

	// Find next free shm_obj where put data
//  TODO WARNING Active wait. If all slots are used,this loops will never stop untill a sa new free slot... find somethings better
/*		int j=logically_free_shm_obj;
fprintf(stderr,"aiolid.c:process_readrqsts(): tableau des shm_objets:");  
		do{
			mutex_lock(shm_mutex[j]);
fprintf(stderr,"%d/%d ",j,shm_objs[j].link);  
			mutex_unlock(shm_mutex[j]);
			j=(j+1)%MAX_SHMNB;
		}
		while(j!=logically_free_shm_obj);
fprintf(stderr,"\n");  
*/
	mutex_lock(shm_mutex[logically_free_shm_obj]);
	while(shm_objs[logically_free_shm_obj].link>0){
		mutex_unlock(shm_mutex[logically_free_shm_obj]);
		logically_free_shm_obj=(logically_free_shm_obj+1)%MAX_SHMNB;
//DIE("aiolid.c:process_readrqsts(): search logically_free_shm_obj:  test %d wich contains link=%d\n",logically_free_shm_obj,shm_objs[logically_free_shm_obj].link);  
		mutex_lock(shm_mutex[logically_free_shm_obj]);
	}
	rreply.shm_index=logically_free_shm_obj;
	logically_free_shm_obj=(logically_free_shm_obj+1)%MAX_SHMNB;
	assert(shm_objs[rreply.shm_index].link==0);
	shm_objs[rreply.shm_index].link=nbrequest; // Now, link > 0 so shm_objs[rreply.shm_index] is protected
	mutex_unlock(shm_mutex[rreply.shm_index]);

	
	if(pthread_mutex_lock(&fdesc_tab[daemonfd].open_close_mutex)!=0){
		DIE("aiolid.c: pthread_mutex_lock Error");
	}
	// Lseek Part, if required
	if(begin_offset!=fdesc_tab[daemonfd].global_position){
		global_result=lseek(daemonfd,begin_offset,SEEK_SET);
		if(global_result<0){
			if(pthread_mutex_unlock(&fdesc_tab[daemonfd].open_close_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_unlock Error");
			}
#if SHOWINFO	
			fprintf(stderr,"aiolid.c:process_readrqsts():Can't set the fd=%d to the whished offset (offset=%ld)\n",daemonfd,begin_offset);
			perror("");
#endif
		}
		else{
	//			fdesc_tab[daemonfd].global_position=offset; // Useless since READ_RQST couldn't be parrallelized currently.
#if SHOWINFO >	2
			fprintf(stderr,"aiolid.c:process_readrqsts():fd=%d set to offset (offset=%ld)\n",daemonfd,begin_offset);
#endif
		}
	}

	if(global_result!=-1){	
		// Read part
		if(length>MAX_COPYBUFFERSZ)
      	DIE("aiolid.c: process_readrqsts(): could not read more than %d bytes", MAX_COPYBUFFERSZ);
		global_result=read(daemonfd,shm_objs[rreply.shm_index].bytes,length);
		if (global_result <0){
			if(pthread_mutex_unlock(&fdesc_tab[daemonfd].open_close_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_unlock Error");
			}
#if SHOWINFO	
			fprintf(stderr,"aiolid.c:process_readrqsts(),read failed (from offset %ld in the file %s (fd=%d))\n",begin_offset,fdesc_tab[daemonfd].pathname,daemonfd);
			perror("");
#endif		
		}
		else{
#if SHOWINFO > 1	
			fprintf(stderr,"aiolid.c:process_readrqsts(),read %d bytes from offset %ld in the file %s (fd=%d)\n",length,begin_offset,fdesc_tab[daemonfd].pathname,daemonfd);
#endif
			fdesc_tab[daemonfd].global_position+=global_result;
			if(pthread_mutex_unlock(&fdesc_tab[daemonfd].open_close_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_unlock Error");
			}
		}
	}
	//fprintf(stderr,"aiolid.c:process_readrqsts(): read ok selected logically_free_shm_obj %d and link set to %d\n",rreply.shm_index,nbrequest);  
	// Send replies to all concerned processus
	// before, if temp_buffer != null, we have to affect each region to each shared buffer.
	// We plan to skip these noisy copies by using a global shared buffer for all processus but TODO :) .
	list_for_each_safe(ptr_position,ptr_copy_position,rrqsts){
		ptr_temp_rrqst=((read_request*)((rqst_cell*)list_entry(ptr_position,rqst_cell,list))->packet);
		if(global_result==length){ // All required bytes have benne retrieve
			// Set seek_start for each proces
				rreply.seek_start=ptr_temp_rrqst->offset-begin_offset;
				rreply.result=ptr_temp_rrqst->length;
		}
		else {
			// TODO Please, check real read bytes for each request. Example eof so wich value will have to return.
		}
		rreply.requestid=ptr_temp_rrqst->requestid;
		rreply.remote_errno=errno; 	// errno should be affect for each reply since memcpy could be failed.
		if(post_reply(process_tab[ptr_temp_rrqst->pid_desc].msgq_replyid, READ_REPLY, &rreply, sizeof(read_reply))==0){
			process_tab[ptr_temp_rrqst->pid_desc].last_ack_request=ptr_temp_rrqst->requestid;
#if SHOWINFO > 2
			fprintf(stderr,"aiolid.c : process_readrqsts() post reply %d of pid_desc=%d ok\n",ptr_temp_rrqst->requestid,ptr_temp_rrqst->pid_desc);
#endif			
		}
#if SHOWINFO
		else{
			fprintf(stderr,"aiolid.c : process_readrqsts() post reply of pid_desc=%d error\n",ptr_temp_rrqst->pid_desc);
		}
#endif			
	}
}

void process_writerqst (write_request *wrqst){
	int result =-1 ;
#if SHOWINFO 		  
	fprintf(stderr,"write request, we have to implement the related code\n");
#endif
	post_genericreply(process_tab[wrqst->pid_desc].msgq_replyid, wrqst->requestid, WRITE_REPLY, result, errno);
}
void process_writerqsts(rqst_list* wrqsts){
	//int result =-1 ;
#if SHOWINFO		  
	fprintf(stderr,"aiolid.c: process_writerqsts(): write requests, we have to implement the related code\n");
#endif
	// TODO send reply to all concerned processes	
//	post_genericreply(process_tab[wrqst->pid_desc].msgq_replyid, wrqst->requestid, WRITE_REPLY, result, errno);
}

void process_initrqst(init_request *irqst){
	int pid_desc = add_process(irqst->pid,irqst->msgq_replykey,irqst->shm_name);
	if(pid_desc<0){
#if SHOWINFO
		fprintf(stderr,"aiolid.c: process_initrqst(): Can't add the processus with pid=%d (errno=%d)\n",irqst->pid,errno);	
//		perror(); //TODO fix the way to manage new error value look at man strerror.
#endif
		// The message queue used for send reply may not be set up (add_process has returned -1), so we have to do it ;)
		
		post_genericreply(init_reply_msgqueue(irqst->msgq_replykey), irqst->requestid, INIT_REPLY, pid_desc, errno);  // Here result is equal to -1.
	}
	else{
#if SHOWINFO > 1 
		fprintf(stderr,"aiolid.c: process_initrqst(): Add a new process: pid_desc = %d and reply msgqueue = %d\n",pid_desc,process_tab[pid_desc].msgq_replyid);	
#endif
		// Here, we could say that process_tab[pid_desc].msgq_replykey is the right ipc message queue to send replies.
		// Actually, any other request could be send to aioli daemon before INIT_RQST finishes. 
		// So, that means any FINALIZE_RQST requests could be send from any thread belonging to process_tab[pid_desc].pid and process_tab[pid_desc].pid hasn't been set to -1.
		// In the case of a segmentation fault from process_tab[pid_desc].pid, post_genericreply will be return -1 and we don't take care about the related client processus.
		if(post_genericreply(process_tab[pid_desc].msgq_replyid, irqst->requestid, INIT_REPLY, pid_desc, errno)>=0)  // Send ack to processus processus_tab[pid_desc].pid
			process_tab[pid_desc].last_ack_request=irqst->requestid;
	}
}


void process_finalrqst(finalize_request *frqst){
	int msgq_replyid =-1;
	// In the case of EBAD_PIDDESC, we don't reply since we couldn't :)
	if((frqst->pid_desc<0) && (frqst->pid_desc>MAX_AGGREGATED_PROCESSES)){	
#if SHOWINFO	
		fprintf(stderr,"aiolid.c: process_finalrqst(), request arguments : bad pid descripteur %d\n",frqst->pid_desc);
#endif
	 	errno= EBAD_PIDDESC;
	}
	else if(process_tab[frqst->pid_desc].pid!=frqst->pid){	
#if SHOWINFO	
		fprintf(stderr,"aiolid.c: process_finalrqst(), request arguments : bad processus id (pid=%d) for pid_desc=%d (waiting for pid=%d)\n",frqst->pid,frqst->pid_desc,process_tab[frqst->pid_desc].pid);
#endif
	 	errno= EBAD_PIDDESC;
	}
	else{
		// Note : remove_process returns msg_reply id, -1 otherwise
		msgq_replyid=process_tab[frqst->pid_desc].msgq_replyid;
		int result = remove_process(frqst->pid_desc,frqst->pid);
#if SHOWINFO 
		if(result<0){
			fprintf(stderr,"aiolid.c: process_finalrqst():  Can't remove process with pid=%d (pid_desc=%d, errno=%d)\n",frqst->pid,frqst->pid_desc,errno);	
	//		perror(); //TODO fix the way to manage new error value.
		}
#endif
#if SHOWINFO > 1		
		else{
			fprintf(stderr,"aiolid.c: process_finalrqst():  Remove process with pid=%d (pid_desc=%d)\n",frqst->pid,frqst->pid_desc);	
		}
#endif
		post_genericreply(msgq_replyid, frqst->requestid, FINAL_REPLY, result, errno);  // 'result' is equal to msg queue id.
		// We don't store last_ack_request since processus has been removed ;) !
	}
	
}

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// PROCESS AND FILE MANAGEMENT PARTS
//
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
void init_tabs(void){
	
	int i,j;
	
	// Process info :
	//processes_mutex=PTHREAD_MUTEX_INITIALIZER; doesn't compile :(.
	if(pthread_mutex_init(&processes_mutex,NULL)!= 0){
#if SHOWINFO
		   fprintf(stderr,"aiolid.c: init_tabs(): Can't initialize processes_mutex : pthread_mutex_init Error\n");
			perror("");
#endif
	}		

	//	pthread_mutex_lock(&processes_mutex); 					// Useless since a this moment there is only one thread but....maybe one day ;)

	for (i=0 ; i < MAX_AGGREGATED_PROCESSES ; i++ ){
		process_tab[i].pid = -1; 
		process_tab[i].msgq_replyid = -1; 
		process_tab[i].nb_file = 0; 
		for (j=0;j<MAX_FILEOPEN;j++){
			process_tab[i].iob[j].daemonfd=-1; 				// Indeed, daemonfd=-1 means that the value is not used
		}
		process_tab[i].last_ack_request = -1; 
 	}
	//	pthread_mutex_unlock(&processes_mutex); 				// As above useless ;)
	
	// File descriptor info : 
	for (i=0 ; i < MAX_FILEOPEN ; i++ ){
		fdesc_tab[i].nb_owners=0;
		for (j=0;j<MAX_FILEOPEN;j++){
			fdesc_tab[i].owners[j]=-1;							// Note : owners[j] enables to retrieve directly the related userfd for the processus j, -1 means that there isn't !  				
		}
		fdesc_tab[i].pathname[0]='\0';
		fdesc_tab[i].global_position=0;
		fdesc_tab[i].flags=0;
		fdesc_tab[i].mode=0;
		//fdesc_tab[i].open_close_mutex=PTHREAD_MUTEX_INITIALIZER; // As above for processes_mutex, doesn't compile :((
		if(pthread_mutex_init(&fdesc_tab[i].open_close_mutex,NULL) != 0){
#if SHOWINFO
		   fprintf(stderr,"aiolid.c: init_tabs(): Can't initialize fdesc_tab[%d].open_close_mutex : pthread_mutex_init Error\n",i);
			perror("");
#endif
		}
	}
}

// WARNINGS : each parameter should be right !!!
// for performance, anything is check inside this function
int add_process(pid_t pid, key_t msgq_replykey, const char* shm_name){

	int pid_desc = -1 ; // The related position in the process_tab (this value is returned to the calling process). 
	int i;
	if(pthread_mutex_lock(&processes_mutex)!=0){
		DIE("aiolid.c: pthread_mutex_lock Error");
	} 					
	// First, find a free slot
  	for (i=0 ; i < MAX_AGGREGATED_PROCESSES ; i++ ){
		if (process_tab[i].pid<0){ 
			pid_desc = i;
			break;
		}
 	}
	
	if (pid_desc <0){ // No free slot
		errno=ENPROCESS; // TODO Have to define ENPROCESS error which means the maximum
				 			 //  number of process has been reached (similar to ENFILE ;), man 2 open).
		// TODO in order to improve our library, we have to send an alive request to each client and remove these will not respond. In this way, we do something similar
		//  to a garbage collector to skip old bad client.
		
		if(pthread_mutex_unlock(&processes_mutex)!=0){
			DIE("aiolid.c: pthread_mutex_unlock Error");
		}
		return -1;
	}

	process_tab[pid_desc].pid = pid ; 
	if(pthread_mutex_unlock(&processes_mutex)!=0){ 	// At this time, process_tab[pid_desc] is protect by process_tab[pid_desc]>=0	
		DIE("aiolid.c: pthread_mutex_unlock Error");
	}
		
	// Initialize the found slot 
	process_tab[pid_desc].msgq_replyid=init_reply_msgqueue(msgq_replykey); 
#if SHOWINFO > 2		  
	fprintf(stderr,"aiolid.c : add_process() : create a \"reply\" message queue (id=%d) for processus pid=%d, pid_desc=%d\n",process_tab[pid_desc].msgq_replyid, pid, pid_desc);	
#endif
//	process_tab[pid_desc].nb_file = 0 ;  
	process_tab[pid_desc].nb_file = 3 ;   //TODO take into account stdin stdout stderr
	
	return pid_desc ; 
}

// Return the reply message queue id otherwise -1. 
int remove_process(int pid_desc,pid_t pid){
	int tmp_msgq_replyid=-1;

	if(((pid_desc<0) && (pid_desc>MAX_AGGREGATED_PROCESSES)) && (process_tab[pid_desc].pid!=pid)){
#if SHOWINFO  
	fprintf(stderr,"aiolid.c:remove_process(): the pid_desc %d has a different pid value!(set errno to E_BAD_PIDDESC)\n",pid_desc);
#endif
		errno=EBAD_PIDDESC; 
		return -1;
	}
	
	//  We have to check if all file desc belong to this desc_pid, have been closed before to delete it !!!
	int i;
#if SHOWINFO > 2
	fprintf(stderr,"aiolid.c:remove_process(): There is/are %d file(s) opened for pid_desc=%d\n",process_tab[pid_desc].nb_file,pid_desc);
#endif	
	for(i=0;i<MAX_FILEOPEN /*&& process_tab[pid_desc].nb_file >0*/ ;i++){ //TODO Find a clean way to only remove the open file !!! (stdin,stderr,stdout problems)
		if (process_tab[pid_desc].iob[i].daemonfd>=0){
	 	//	close_file_desc(pid_desc,i,process_tab[pid_desc].iob[i].daemonfd);	
			process_tab[pid_desc].iob[i].daemonfd=-1;
			process_tab[pid_desc].nb_file--;
		}
	}
	tmp_msgq_replyid=process_tab[pid_desc].msgq_replyid;
	process_tab[pid_desc].msgq_replyid=-1;
	process_tab[pid_desc].pid=-1; // WARNINGS, That means the process_tab[pid_desc] is now free and could be taken by another processus

	return tmp_msgq_replyid;	
}

// WARNINGS : each parameter should be right !!!
// for performance, anything is check inside this function
// Return : daemonfd, -1 otherwise
int add_file(int pid_desc,int userfd, const char* pathname, int daemonfd){

	// update fdesc_tab infos
	if(fdesc_tab[daemonfd].nb_owners>=MAX_AGGREGATED_PROCESSES){
#if SHOWINFO		  
		fprintf(stderr,"aiolid.c:add_file(): Error, can't add a new 'owner' to the file %s\n",fdesc_tab[daemonfd].pathname);
#endif
		// TODO check How to handle new error value.
		errno=EN_AGGREGATEDPROCESSES;
		return -1;
	}
	else if(fdesc_tab[daemonfd].nb_owners>0){ // you just have to affect to fdesc_tab for the owner desc_pid the relarted userfd
	   fdesc_tab[daemonfd].owners[pid_desc]=userfd; // To affect userfd to the owners tab provided a way to retrieve directly the related user file descriptor.
																	// For example : The related user file descriptor for the client processus desc_pid an the file handle by daemonfd
																	//  is fdesc_tab[daemon].owners[pid_desc].
																	//  This trick will be really performant to manage the global shared ptr concept as in MPI I/O Standard.
		fdesc_tab[daemonfd].nb_owners++;
#if SHOWINFO > 2 	  
		fprintf(stderr,"aiolid.c add_file(): Add a new owner on file called %s\n",pathname);
#endif

	}
	else{ // this is the first desc_pid to open this file so...
#if SHOWINFO > 2
		fprintf(stderr,"aiolid.c add_file(): Add a new file called %s\n",pathname);
#endif
		strcpy(fdesc_tab[daemonfd].pathname,pathname); // TODO A way to forbid the overflow could be to use strncpy with n=MAX_FILENAMESZ
//		fdesc_tab[daemonfd].global_position=0; // TODO Should be just after open system call
		fdesc_tab[daemonfd].owners[pid_desc]=userfd;
		fdesc_tab[daemonfd].nb_owners++;
	}
	return daemonfd;
}

// return number of remain owner for this file, otherwise -1;
int remove_file(int pid_desc, int userfd, int daemonfd){

	if(fdesc_tab[daemonfd].nb_owners==0){
#if SHOWINFO
		fprintf(stderr,"aiolid.c:remove_file(): Something seems wrong: This file seems to be not used (nb_owners=0)\n");
#endif
		// TODO Please check what does EBADF represent ?
		errno=EBADF;
	 	return -1;
	}
	else if(fdesc_tab[daemonfd].owners[pid_desc]==-1){
#if SHOWINFO
		fprintf(stderr,"aiolid.c:remove_file(): Something seems wrong: This file seems to be not used by the process with desc_pid=%d.\n",pid_desc);
#endif
		errno=EBAD_PIDDESC;
	 	return -1;
	}
	else if(fdesc_tab[daemonfd].owners[pid_desc]!=userfd){
#if SHOWINFO
	fprintf(stderr,"aiolid.c:remove_file(): Something seems wrong: This asked file doesn't correspond to the given file descriptor (desc_pid=%d,userfd=%d and related value=%d\n"
						 ,pid_desc,userfd,fdesc_tab[daemonfd].owners[pid_desc]);
#endif
		errno=EBADF;
	 	return -1;
	}
	
	fdesc_tab[daemonfd].owners[pid_desc]=-1;
	fdesc_tab[daemonfd].nb_owners--;	
	return fdesc_tab[daemonfd].nb_owners ; 
}	

// Return the daemon file desciptor.
// WARNINGS : any parameters have been check (overflow, ...), so have to ;)
int open_file_desc(int pid_desc, const char* pathname, int flags, int mode, int userfd){ 
	// TODO Define the default behaviour when flags and mode are different from a file already open. Example, O_READ vs O_WRITE etc...
	// Currently, if flags = O_READ, that's right otherwise undefined.... Maybe find a way to compare flags and mode value.
	
  	int daemonfd=-1;
	int selected_daemonfd=-1;  // We need a temporary daemon file descriptor, actually the real daemonfd will be returned by add_file()
										// and in order to call this function, we have to hold fdesc_tab[selected_daemonfd].open_close_mutex.
										// In the case of we store in only one variable daemonfd, we could try to unlock fdesc_tab[-1].open_close_mutex since add_file() could be return -1.
	int i=0;

	//First, check if the file is already open with same flags and same mode.
	for(i=0; i<MAX_FILEOPEN ; i++){
		if(pthread_mutex_lock(&fdesc_tab[i].open_close_mutex)!=0){
			DIE("aiolid.c: pthread_mutex_lock Error");
		}
		if((fdesc_tab[i].nb_owners!=0)&&(strcmp(fdesc_tab[i].pathname,pathname)==0)){
			selected_daemonfd=i;
		// Actually, here we have to add a flags and mode compares functions.
#if SHOWINFO > 2
			fprintf(stderr,"aiolid.c:open_file_desc(): file %s already open so just return file descirptor %d.\n",pathname,daemonfd);				  
#endif 				  
			break;
		}
		else
			if(pthread_mutex_unlock(&fdesc_tab[i].open_close_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
	}

// TODO Check add_file position call
	if(selected_daemonfd<0){ // The file have to be open and no mutex is hold
 		selected_daemonfd=open(pathname,flags,mode);
#if SHOWINFO > 2
		fprintf(stderr,"aiolid.c:open_file_desc(): make a \"real\" open call on file %s and retrieve file descirptor %d.\n",pathname,selected_daemonfd);				  
#endif		

		if(selected_daemonfd>=0) {
			if(pthread_mutex_lock(&fdesc_tab[selected_daemonfd].open_close_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
			daemonfd=add_file(pid_desc,userfd,pathname,selected_daemonfd); 
			if(daemonfd<0)
				close(selected_daemonfd);	  
			if(pthread_mutex_unlock(&fdesc_tab[selected_daemonfd].open_close_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_unlock Error");
			}
		}
	}
	else{ // The file is already opened, so we already hold fdesc_tab[daemonfd].open_close_mutex
		daemonfd=add_file(pid_desc,userfd,pathname,selected_daemonfd); 
		if(pthread_mutex_unlock(&fdesc_tab[selected_daemonfd].open_close_mutex)!=0){
			DIE("aiolid.c: pthread_mutex_lock Error");
		}
	}
	return daemonfd; // if everything is right : selected_daemonfd ==n daemonfd;
}

// Return the number of remain owners for the file (daemonfd)
int close_file_desc(int pid_desc, int userfd, int daemonfd){
	int result;
	
	if(pthread_mutex_lock(&fdesc_tab[daemonfd].open_close_mutex)!=0){
		DIE("aiolid.c: pthread_mutex_lock Error");
	}
	// remove_file returns the number of remaining references on the file handled by daemonfd.
	result=remove_file(pid_desc, userfd, daemonfd);

	if (result==0){ // It was the last reference, so we have to "really" close it ;)
	   result=close(daemonfd);
		if(result<0){
#if SHOWINFO
			fprintf(stderr,"aiolid.c:close_file_desc(): Error, file %s can't be closed\n",fdesc_tab[daemonfd].pathname);
			perror("");
#endif
		}
		else {
#if SHOWINFO > 2
			fprintf(stderr,"aiolid.c:close_file_desc(): file %s has been closed (by a \"close\" call)\n",fdesc_tab[daemonfd].pathname);
#endif
#if SHOWAGGREGATED
			fprintf(stderr,"aiolid.c:close_file_desc(): file %s has been closed (%lu READ_REQ (max_agglomeration=%d)and %lu WRITE_REQ request(s) have been aggregated)\n",fdesc_tab[daemonfd].pathname,
								 fdesc_request[daemonfd].readaggregatedrqst_nb,fdesc_request[daemonfd].maxread_aggolmeration,fdesc_request[daemonfd].writeaggregatedrqst_nb);
			fdesc_request[daemonfd].readaggregatedrqst_nb=0;	
			fdesc_request[daemonfd].maxread_aggolmeration=0;	
			fdesc_request[daemonfd].writeaggregatedrqst_nb=0;	
#endif 			  
			
			INIT_LIST_HEAD(&fdesc_request[daemonfd].readrqst_list);
			fdesc_request[daemonfd].readrqst_nb=0;
			fdesc_request[daemonfd].aggregate_readprocess=0;
			INIT_LIST_HEAD(&fdesc_request[daemonfd].writerqst_list);
			fdesc_request[daemonfd].writerqst_nb=0;
			fdesc_request[daemonfd].aggregate_writeprocess=0;

			fdesc_tab[daemonfd].pathname[0]='\0';
			fdesc_tab[daemonfd].global_position=0;
			fdesc_tab[daemonfd].flags=0;
			fdesc_tab[daemonfd].mode=0;
		}
	}
	if(pthread_mutex_unlock(&fdesc_tab[daemonfd].open_close_mutex)!=0){
		DIE("aiolid.c: pthread_mutex_unlock Error");
	}
	return result;
}

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// LINKED LIST PART
// 
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 

// Should be called before to create aioli thread !!!
pthread_mutexattr_t rqsts_mutexattr;
void init_lists(){
	int i;
	if(pthread_mutexattr_init(&rqsts_mutexattr) != 0)
		perror("mutexattr_settype Error\n");
	if(pthread_mutexattr_settype(&rqsts_mutexattr,PTHREAD_MUTEX_ERRORCHECK_NP) != 0)
		perror("mutexattr_settype Error\n");

	if(pthread_mutex_init(&rqsts_mutex,&rqsts_mutexattr) != 0){
#if SHOWINFO 
		fprintf(stderr,"aiolid.c: init_lists(): Can't initialize rqsts_mutex : pthread_mutex_init Error\n");
		perror("");
#endif
	}	
 	if(pthread_cond_init(&rqst_cond_free,NULL) != 0){
#if SHOWINFO
		fprintf(stderr,"aiolid.c: init_lists(): Can't initialize rqst_cond_free : pthread_cond_init Error\n");
		perror("");
#endif
	}	
	if(pthread_cond_init(&rqst_cond_filled,NULL) != 0){
#if SHOWINFO
		fprintf(stderr,"aiolid.c: init_lists(): Can't initialize rqst_cond_filled : pthread_cond_init Error\n");
		perror("");
#endif
	}	

	// Free list		  
	INIT_LIST_HEAD(&free_list);	
	for (i=0;i<MAX_RQSTNB;i++){
		rqst_cells[i].state=FREE;
		rqst_cells[i].time_position=0;
		list_add_tail(&(rqst_cells[i].list),&free_list);
	}

	// Time order list
	INIT_LIST_HEAD(&time_order_list);	

	// Sorted Lists by request type
	for (i=0;i<NB_REQ_TYPE-2;i++){
		INIT_LIST_HEAD(&(lists[i]));	
	}

	// Read and write List by file descriptor
	for (i=0;i<MAX_FILEOPEN;i++){
		INIT_LIST_HEAD(&fdesc_request[i].readrqst_list);
		fdesc_request[i].readrqst_nb=0;
		fdesc_request[i].aggregate_readprocess=0;
#if SHOWAGGREGATED
		fdesc_request[i].readaggregatedrqst_nb=0;	// Just to know how many request have been aggregated for this file (to first open call untill last close call).	
		fdesc_request[i].maxread_aggolmeration=0;	
#endif 			  
		INIT_LIST_HEAD(&fdesc_request[i].writerqst_list);
		fdesc_request[i].writerqst_nb=0;
		fdesc_request[i].aggregate_writeprocess=0;
#if SHOWAGGREGATED
		fdesc_request[i].writeaggregatedrqst_nb=0;	// Just to know how many request have been aggregated for this file (to first open call untill last close call).	
#endif 			  
	}
}

void add_request (int type, void *request){

	rqst_cell *ptr_rqst_new;

	struct list_head* ptr_position;			// Required in the list_for_each_safe to put the request at the right position in the related list
	struct list_head* ptr_copy_position; 	// The safety reference to position (as above, required in the list_for_each_safe macro).
	rqst_cell *ptr_rqst_temp;					// Required for list iteration

	// Check the type value;
	assert(type>=0);
  	assert(type<NB_REQ_TYPE);
	
	// First, Find a free cell
	if(pthread_mutex_lock(&rqsts_mutex)!=0){
		DIE("aiolid.c: pthread_mutex_lock Error");
	}
	while(list_empty(&free_list)){
		if(pthread_cond_wait(&rqst_cond_free, &rqsts_mutex)!=0){
			DIE("aiolid.c: pthread_cond_wait Error");
		}
	} 
	// If we are here, then a cell is free. So just take the first one ;) and remove it from the free_list
	ptr_rqst_new=list_entry(free_list.next,rqst_cell,list);
	list_del(&(ptr_rqst_new->list));

	if(pthread_mutex_unlock(&rqsts_mutex)!=0){
		DIE("aiolid.c: pthread_mutex_unlock Error");
	}
	assert(ptr_rqst_new->state==FREE);
	
	// Second, Add	in the related list switch type
	ptr_rqst_new->state=RECV;
	ptr_rqst_new->type=type;

	switch(ptr_rqst_new->type){
		// For INIT_REQ, OPEN_REQ, CLOSE_REQ, FINALIZE_REQ the code is similar except the last parameter of the memcpy call ! 
		case INIT_REQ :
			memcpy(ptr_rqst_new->packet,request,sizeof(init_request));
			if(pthread_mutex_lock(&rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
			list_add_tail(&(ptr_rqst_new->list),&(lists[type]));
#if SHOWINFO > 1
			fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c:add_request(): Add an INIT_REQ\n"END_COLOR_CHARACTERS);	
#endif 				  
			break; 
		case OPEN_REQ :
			memcpy(ptr_rqst_new->packet,request,sizeof(open_request));
			if(pthread_mutex_lock(&rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
			list_add_tail(&(ptr_rqst_new->list),&(lists[type]));
#if SHOWINFO > 1
			fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c:add_request(): Add an OPEN_REQ\n"END_COLOR_CHARACTERS);	
#endif 				  
			break; 
		case CLOSE_REQ : 
			memcpy(ptr_rqst_new->packet,request,sizeof(close_request));
			if(pthread_mutex_lock(&rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
			list_add_tail(&(ptr_rqst_new->list),&(lists[type]));
#if SHOWINFO > 1
			fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c:add_request(): Add a CLOSE_REQ\n"END_COLOR_CHARACTERS);	
#endif 				  
			break; 
		case FINAL_REQ : 
			memcpy(ptr_rqst_new->packet,request,sizeof(finalize_request));
			if(pthread_mutex_lock(&rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
			list_add_tail(&(ptr_rqst_new->list),&(lists[type]));
#if SHOWINFO > 1
			fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c:add_request(): Add a FINALIZE_REQ\n"END_COLOR_CHARACTERS);	
#endif 				  
			break; 
		// For READ_REQ and WRITE_REQ, we sort request by file descriptor in an ascending order : 
		// 1./ Find the related file descriptor
		// 2./ Add at the right position in the related list.
		case READ_REQ : 
			memcpy(ptr_rqst_new->packet,request,sizeof(read_request));
			if(pthread_mutex_lock(&rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
#if SHOWINFO > 2
			fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c:add_request(): READ_REQ for daemonfd=%d\n"END_COLOR_CHARACTERS,((read_request*)request)->daemonfd);	

			if(!list_empty(&(fdesc_request[((read_request*)request)->daemonfd].readrqst_list)))
				fprintf(stderr,GREEN_COLOR_CHARACTERS"aiolid.c : add_request(): at least one READ_REQ in READ queue\n"END_COLOR_CHARACTERS);								
			else	
				fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c : add_request(): no  READ_REQ already present in queue\n"END_COLOR_CHARACTERS);								 
#endif
			list_for_each_safe(ptr_position,ptr_copy_position,&(fdesc_request[((read_request*)request)->daemonfd].readrqst_list)){
				ptr_rqst_temp=list_entry(ptr_position,rqst_cell,list);
				if ( ( ((read_request*)ptr_rqst_new->packet)->offset < ((read_request*)ptr_rqst_temp->packet)->offset )||
					 ( ( ((read_request*)ptr_rqst_new->packet)->offset == ((read_request*)ptr_rqst_temp->packet)->offset ) && ( ((read_request*)ptr_rqst_new->packet)->length < ((read_request*)ptr_rqst_temp->packet)->length))) 
						  break;
			}

			list_add(&(ptr_rqst_new->list),ptr_position->prev);
		  	fdesc_request[((read_request*)request)->daemonfd].readrqst_nb++;
#if SHOWINFO > 1
			fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c:add_request(): Add a READ_REQ (list contains %d request(s))\n"END_COLOR_CHARACTERS,
								 fdesc_request[((read_request*)request)->daemonfd].readrqst_nb);
#endif 				  
			break; 
		case WRITE_REQ : 
			memcpy(ptr_rqst_new->packet,request,sizeof(write_request));
			if(pthread_mutex_lock(&rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
		
			list_for_each_safe(ptr_position,ptr_copy_position,&(fdesc_request[((write_request*)request)->daemonfd].writerqst_list)){
				ptr_rqst_temp=list_entry(ptr_position,rqst_cell,list);
				if ( ( ((write_request*)ptr_rqst_new->packet)->offset < ((write_request*)ptr_rqst_temp->packet)->offset )||
					 ( ( ((write_request*)ptr_rqst_new->packet)->offset == ((write_request*)ptr_rqst_temp->packet)->offset ) && ( ((write_request*)ptr_rqst_new->packet)->length < ((write_request*)ptr_rqst_temp->packet)->length))) 
						  break;
			}	
			list_add(&(ptr_rqst_new->list),ptr_position->prev); 
		  	fdesc_request[((write_request*)request)->daemonfd].writerqst_nb++;
#if SHOWINFO > 1
			fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c:add_request(): Add a WRITE_REQ\n"END_COLOR_CHARACTERS);	
#endif 				  
			break; 
		default : 
			DIE("aiolid.c:add_request():Bad Request or unmanage type %d\n",type);	
		// In the other case (NOT DIE !) remeber to take the mutex to be coherent with the next unlock
	   //	if(pthread_mutex_lock(&rqsts_mutex)!=0){
		//		DIE("aiolid.c: pthread_mutex_lock Error");
	   //	}

			break; 
	}	

	// Finally, add request in time order list
	
	if(!list_empty(&time_order_list)){
		list_add_tail(&(ptr_rqst_new->time_list),&time_order_list);
		ptr_rqst_new->time_position=(((rqst_cell*)list_entry(ptr_rqst_new->time_list.prev,rqst_cell,time_list))->time_position)+1; // WARNINGS : time_position must be an uint_64 to avoid rollover pb!
	}
	else{
		ptr_rqst_new->time_position=0; 
		list_add_tail(&(ptr_rqst_new->time_list),&time_order_list);
	}
	ptr_rqst_new->state=FILLED;
#if SHOWINFO > 1
			fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c:add_request(): add a request in time_order_list at the position: %llu\n"END_COLOR_CHARACTERS,ptr_rqst_new->time_position);	
#endif 				  
	
	// Wake up I/O thread
	if(pthread_cond_signal(&rqst_cond_filled)!=0){
		DIE("aiolid.c: pthread_cond_signal Error");
	}

	if(pthread_mutex_unlock(&rqsts_mutex)!=0){
		DIE("aiolid.c: pthread_mutex_unlock Error");
	}
}
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// THREAD PART
//
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
// Global values : 


// Handles the common IPC message queue : Grabs IPC packets and store them in tab_list[] (see aiolid.h)
void* receiver_thr(void *dummy) {
	msg_request msg;

	//Dprintf(D_IOD, "receiver_thr(pid=%d): started as uid=%d euid=%d gid=%d egid=%d",getpid(), getuid(), geteuid(), getgid(), getegid());

#if HAS_PRIO
	if (-1 == setpriority(PRIO_PROCESS, getpid(), PRIO_RECEIVER_THR)) {
//		Dprintf(L_FATAL, "aiolid.c:receiver_thr():setpriority(receiver_thr,%d) failed errno=%d", PRIO_UDP_THR, errno);
//		exit(1);
		fprintf(stderr,"aiolid.c:receiver_thr():setpriority(receiver_thr,%d) failed errno=%d\n", PRIO_RECEIVER_THR,errno);
	}

	/* TODO : Understand below code !!! */
   /* 
	if (-1 == setregid(user->pw_gid, user->pw_gid)) {
		Dprintf(L_FATAL, "cannot change to %s uid/gid", user->pw_name);
		exit(1);
	}
	//	Dprintf(D_IOD, "UDP_thr(pid=%d): running as uid=%d euid=%d gid=%d egid=%d", getpid(), getuid(), geteuid(), getgid(), getegid());
 */
#endif /* HAS_PRIO */
	
	//----------------------------------------------------------

	signal(SIGINT, (__sighandler_t) clean_ipc); // for CTRL+T
	signal(SIGQUIT, (__sighandler_t) clean_ipc);
//	signal(SIGKILL, (__sighandler_t) clean_ipc); // KILL and STOP could'nt be catched
//	signal(SIGSTOP, (__sighandler_t) clean_ipc); // KILL and STOP couldn't be catchec
	signal(SIGTERM, (__sighandler_t) clean_ipc);
	signal(SIGSEGV, (__sighandler_t) clean_ipc);
	
	if(atexit(clean_ipc)!=0)
		DIE("cannot set \"clean_ipc\" as an exit function\n");



	while (1) {
		msg.type=-1;
	
		// First, retrieve a request from the IPC message queue.
#if SHOWINFO > 1
		fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c:receiver_thr():waiting for request\n"END_COLOR_CHARACTERS);
#endif		
		if (msgrcv(msgqid, &msg, MAX_PACKETSZ, 0, 0) < 0){ 
			DIE("aiolid.c:get_request():msgrcv error");
		}
		// Then, add new request in related list
#if SHOWINFO > 3
		fprintf(stderr,RED_COLOR_CHARACTERS"aiolid.c:receiver_thr():request received, we decrement type as ipc message queue required a type > 0 and INIT_REQ=0\n"
							 "Note : type has been incremented just before to be post by optimizediolib\n"END_COLOR_CHARACTERS);
#endif
		msg.type--;		 // remember : msg.type-- since msg.type have to be greater than O and INIT_REQ==0
		add_request(msg.type,msg.packet);
	} // While (1)	;

}

// Handles READ/WRITE/PING requests
// Ideally priority should be the lowest
void* IO_thr(void *dummy) {
	int index;
	
	rqst_cell *ptr_rqst_selected=NULL;
	rqst_cell *ptr_rqst_temp=NULL;
	// Below variables are required to build the aggregated request
	rqst_cell *ptr_rqst_backward=NULL; 	// Required to build the temp_list
	rqst_cell *ptr_rqst_forward=NULL; 		// Required to build the temp_list : Store the request @ used to start the aggregate process.
	rqst_list temp_list;					// The linked list 'temp_list' stores the aggregate request in a ascending offset order
	rqst_list *ptr_related_list_head=NULL;		// A reference to the head of the related list is required to parse it ;)
	u_int64_t time_position_condition=-1;	
	// Required to know if we could concatenate request.
  	//	Indeed, we should'nt aggregate two read requests if there is a write beetween each other (from time point of view)
	//	If we want an UNIX coherence, we have to execute like this way : first read then the write and finally last read.
	// This is similar for write aggregating process.
	// TODO rollover pb persist for uint_64 type ( :~~~( but have no more time to find a clean solution ;) )
												
	struct list_head *ptr_position; 			// Required for list_for_each_safe
	struct list_head *ptr_copy_position;	// Required for list_for_each_safe - Sure something better could easily do but .... ;)

	// Returns the earliest timestamp for given list.
	// This function is required during the aggregation process. Please see below (match next 'find_time_position' ;) )	
	u_int64_t find_time_position(rqst_list* time){
		struct list_head *ptr_position; 			// Required for list_for_each_safe
		struct list_head *ptr_copy_position;	// Required for list_for_each_safe
		rqst_cell *ptr_rqst_temp;
	//	u_int64_t temp_position=(1<<(sizeof(u_int64_t)*8))-1;
		u_int64_t temp_position=~0;
#if SHOWINFO > 2
		fprintf(stderr,"aiolid.c: find_time_position() gonna start\n");
#endif		
		if(!list_empty(time)){
			list_for_each_safe(ptr_position,ptr_copy_position,time){
				ptr_rqst_temp=list_entry(ptr_position,rqst_cell,time_list);
				if(temp_position <ptr_rqst_temp->time_position);
					temp_position=ptr_rqst_temp->time_position;
			}
		}
		return temp_position;	
	}
	struct timeval endtime;
	struct timeval thinktime;
	thinktime.tv_sec=0;
	thinktime.tv_usec=3000;
  	struct timespec waittime;
	int retcode=0;
	// WARNINGS, this method doesn't check any overflow pbs. Still Phd Guy so still no time ;)
	void timeadd(struct timeval *t1, struct timeval *t2){
		t1->tv_sec=t1->tv_sec+t2->tv_sec+((t1->tv_usec+t2->tv_usec)/1000000);
		t1->tv_usec=(t1->tv_usec+t2->tv_usec)%1000000;
	}

//	Dprintf(D_IOD, "IO_thr(pid=%d): started as uid=%d euid=%d gid=%d egid=%d",getpid(), getuid(), geteuid(), getgid(), getegid());

#if HAS_PRIO
/*	if (-1 == seteuid(0)) {
		Dprintf(L_FATAL, "cannot change to uid 0");
		exit(1);
	}
*/
	if (-1 == setpriority(PRIO_PROCESS, getpid(), PRIO_IO_THR)) {
		fprintf(stderr,"aiolid.c : IO_thr() : setpriority(IO_thr,%d) failed errno=%d", PRIO_IO_THR, errno);
	}
/*	if (-1 == setregid(user->pw_gid, user->pw_gid)) {
		Dprintf(L_FATAL, "cannot change to %s gid", user->pw_name);
		exit(1);
	}
	if (-1 == setreuid(user->pw_uid, user->pw_uid)) {
		Dprintf(L_FATAL, "cannot change to %s uid", user->pw_name);
		exit(1);
	}
	
	Dprintf(D_IOD, "IO_thr(pid=%d): running as uid=%d euid=%d gid=%d egid=%d",
	     getpid(), getuid(), geteuid(), getgid(), getegid());*/
#endif /* HAS_PRIO */
	

	while (1) {

		// First, Find a filled cell

#if SHOWINFO > 2
		fprintf(stderr,"aiolid.c : IO_thr() : Asked lock on mutex rqsts_mutex\n");
#endif		
	   if(pthread_mutex_lock(&rqsts_mutex)!=0){
			DIE("aiolid.c: pthread_mutex_lock Error");
		}
		while(list_empty(&time_order_list)){
			if(pthread_cond_wait(&rqst_cond_filled, &rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_cond_wait Error");
			}
		/* Active Waiting */
	/*	struct timespec activewait;
		activewait.tv_sec=0;
		activewait.tv_nsec=100;
		if(pthread_mutex_unlock(&rqsts_mutex)!=0){
			DIE("aiolid.c: pthread_mutex_lock Error");
		}	
		nanosleep(&activewait,NULL);
		if(pthread_mutex_lock(&rqsts_mutex)!=0){
			DIE("aiolid.c: pthread_mutex_lock Error");
		}	*/
		/* End of active waiting*/
		} 
#if SHOWINFO > 2
		fprintf(stderr,"aiolid.c : IO_thr() : Lock on mutex rqsts_mutex ok\n");
#endif		

		// If we're here then at least 1 cell is marked FILLED
		ptr_rqst_selected=list_entry(time_order_list.next,rqst_cell,time_list);
#if SHOWINFO > 1 
			fprintf(stderr,"aiolid.c:IO_thr(): type of selected request = %ld (gonna start to check if one more important request is present)\n",ptr_rqst_selected->type);	
#endif 				  

		// We have to check if there is one request with a higher priority level.
		index=INIT_REQ; 
		while ((index<ptr_rqst_selected->type)&&(index<READ_REQ)) {
			if (! list_empty(&lists[index]) ){
#if SHOWINFO > 2
			fprintf(stderr,"aiolid.c:IO_thr(): lists[%d] is not empty so we gonna process one of its requests\n",index);	
#endif 				  

				break;
			}
			index++;
		}
#if SHOWINFO > 2
			fprintf(stderr,"aiolid.c:IO_thr(): index=%d and type of selected request = %ld\n",index,ptr_rqst_selected->type);	
#endif 				  
	
		if((index!=ptr_rqst_selected->type) && (index<READ_REQ)){ // A request has a higher priority so we have to process it ;)
			ptr_rqst_selected=list_entry(lists[index].next,rqst_cell,list);
		}
#if SHOWINFO > 1
			fprintf(stderr,"aiolid.c:IO_thr(): type of selected request = %ld \n",ptr_rqst_selected->type);	
#endif 				  
		
		// Here, we have found the right request.
		assert(ptr_rqst_selected->state==FILLED);
		
		if(ptr_rqst_selected->type < READ_REQ){
			// So, remove the selected request from time_order_list and from the related list
			list_del(&(ptr_rqst_selected->time_list));
			list_del(&(ptr_rqst_selected->list));

			ptr_rqst_selected->state=USED;
			if(pthread_mutex_unlock(&rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_unlock Error");
			}
			
			// Have to process it ;)
			switch (ptr_rqst_selected->type){
				// TODO : Why don't try to aggregate multiple OPEN_REQ on the same file ? Think about it ;)
				// I'm not really sure it lead to better performances !!
				case OPEN_REQ : 
					process_openrqst((open_request*)ptr_rqst_selected->packet);
					break ;
				case CLOSE_REQ : 
					process_closerqst((close_request*)ptr_rqst_selected->packet);
					break ;
				case INIT_REQ :
					process_initrqst((init_request*)ptr_rqst_selected->packet); 
					break;
				case FINAL_REQ :
					process_finalrqst((finalize_request*)ptr_rqst_selected->packet);
					break;
				default : 
					DIE("aiolid.c:IO_thr(): get a request with a bad type : %ld undefined",ptr_rqst_selected->type);
			}
		
			//add proccessed request in the free_list	
			if(pthread_mutex_lock(&rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
			
			ptr_rqst_selected->state=FREE;
			list_add(&ptr_rqst_selected->list,&free_list);
#if SHOWINFO > 1
			fprintf(stderr,"aiolid.c:IO_thr(): The selected request (type=%ld) has been processed (a new free cell has been added to the free_list).\n",ptr_rqst_selected->type);	
#endif 				  

			// Wake up receiver thread
			if (pthread_cond_signal(&rqst_cond_free)!=0){
				DIE("aiolid.c: pthread_cond_signal Error");
			}
			if(pthread_mutex_unlock(&rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_unlock Error");
			}
#if SHOWINFO > 2
			fprintf(stderr,"aiolid.c:IO_thr(): wake up other ok\n");	
#endif 				  

		}
		else{ // This is a READ_REQ/WRITE_REQ, so we gonna try to do something better ;)
	  
		 	 // Waiting for aggregated requests
			// TODO only READ_REQ things have been implemented and we have to check this part of code. it is non thread-safe ! since 
				if ((ptr_rqst_selected->type==READ_REQ)&& fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].aggregate_readprocess>0){
					if (-1 == gettimeofday(&endtime, NULL))  
      				DIE("aiolid.c: IO_thr(): could not get start time (gettimeofday() failed errno=%d)\n", errno);
#if SHOWINFO > 2
					else
						fprintf(stderr,"aiolid.c: IO_thr(): endttime: %ld sec and %ld microsec\n",endtime.tv_sec,endtime.tv_usec);
#endif
					timeadd(&endtime,&thinktime);
					waittime.tv_sec=endtime.tv_sec;
					waittime.tv_nsec=endtime.tv_usec*1000;
					if (waittime.tv_nsec >= 1000000000){ // avoid overflow issue. TODO : check
						waittime.tv_nsec -= 1000000000;
						waittime.tv_sec++;
					}
					
					// I already hold  rqsts_mutex
					while((retcode!=ETIMEDOUT)&&(fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].readrqst_nb<fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].aggregate_readprocess)){
						retcode=pthread_cond_timedwait(&rqst_cond_filled, &rqsts_mutex,&waittime);
#if SHOWINFO > 2
						if(retcode==ETIMEDOUT)
							fprintf(stderr,"aiolid.c: IO_thr(): pthread_cond_timedwait timeout (requestid=%d, from pid_desc=%d)\n",((read_request*)ptr_rqst_selected->packet)->requestid,((read_request*)ptr_rqst_selected->packet)->pid_desc);
						else
							fprintf(stderr,"aiolid.c: IO_thr(): pthread_cond_timedwait fulfill\n");
#endif	
					} 
				}
				retcode=0;


				// If a READ_REQ is present and no aggregate process have been started. 
				if ( ((ptr_rqst_selected->type==READ_REQ)&&(fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].readrqst_nb==1))
//								 && 		(fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].aggregate_readprocess==0)
								  ){
						// Simple READ_REQ, so just process_it
					list_del(&(ptr_rqst_selected->time_list));
					list_del(&(ptr_rqst_selected->list));
					fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].readrqst_nb--;
					ptr_rqst_selected->state=USED;
					if(pthread_mutex_unlock(&rqsts_mutex)!=0){
						DIE("aiolid.c: pthread_mutex_unlock Error");
					}
					
					process_readrqst((read_request*)ptr_rqst_selected->packet);
					
					// Add proccessed request(s) in the free_list	
					if(pthread_mutex_lock(&rqsts_mutex)!=0){
						DIE("aiolid.c: pthread_mutex_lock Error");
					}
					ptr_rqst_selected->state=FREE;
					list_add(&(ptr_rqst_selected->list),&free_list);
#if SHOWINFO > 1
			fprintf(stderr,YELLOW_COLOR_CHARACTERS"aiolid.c:IO_thr(): A READ_REQ request has been proccessed. A new free cell has been added to the free_list\n"END_COLOR_CHARACTERS);	
#endif 				  
				}
				else if( ((ptr_rqst_selected->type==WRITE_REQ)&&(fdesc_request[((write_request*)ptr_rqst_selected->packet)->daemonfd].writerqst_nb==1))
//								 && (fdesc_request[((write_request*)ptr_rqst_selected->packet)->daemonfd].aggregate_writeprocess==0) 
							){
				// Simple WRITE_REQ, so just process_it
				
				// First, remove the selected request from time_order_list and from the related list
					list_del(&(ptr_rqst_selected->time_list));
					list_del(&(ptr_rqst_selected->list));
					fdesc_request[((write_request*)ptr_rqst_selected->packet)->daemonfd].writerqst_nb--;
					ptr_rqst_selected->state=USED;
					if(pthread_mutex_unlock(&rqsts_mutex)!=0){
						DIE("aiolid.c: pthread_mutex_unlock Error");
					}
					
					process_writerqst((write_request*)ptr_rqst_selected->packet);
				
					// Add proccessed request(s) in the free_list	
					if(pthread_mutex_lock(&rqsts_mutex)!=0){
						DIE("aiolid.c: pthread_mutex_lock Error");
					}
					ptr_rqst_selected->state=FREE;
					list_add(&(ptr_rqst_selected->list),&free_list);
#if SHOWINFO > 1
					fprintf(stderr,"aiolid.c:IO_thr(): A WRITE_REQ request has been proccessed. A new free cell has been added to the free_list\n");	
#endif 		
				}
				else{ // There are several request so try to "aggregate" as soon as possible

#if SHOWINFO > 1
					fprintf(stderr,"aiolid.c:IO_thr(): READ_REQ or WRITE_REQ (type=%ld)" 
								 "(backward and forward aggregate process gonna start).\n",ptr_rqst_selected->type);	
#endif 				  

				// Find the first request that have to be executed in the 'opposite' list in order to add time constraint in our aggregate model.
				// TODO : This easy time approach has to be improve with Lamport concept when we'll deploy our model to clusters and grids
					if(ptr_rqst_selected->type == READ_REQ){
						ptr_related_list_head = &(fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].readrqst_list);
						time_position_condition=find_time_position(&fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].writerqst_list);
					}
					else{	// WRITE_REQ	
						ptr_related_list_head = &(fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].writerqst_list);
						time_position_condition = find_time_position(&fdesc_request[((write_request*)ptr_rqst_selected->packet)->daemonfd].readrqst_list);
					}
#if SHOWINFO > 2
					fprintf(stderr,"aiolid.c:IO_thr(): find time constraint = %llu \n",time_position_condition);	
#endif 	
			// Add naive collective junction part
			// Parse the whole request stored in related read list.
			// TODO : Notes : Only READ_REQ case has been implemented.... 
					list_for_each_safe(ptr_position,ptr_copy_position,ptr_related_list_head){
						ptr_rqst_temp=list_entry(ptr_position,rqst_cell,list);
						if ( (((read_request*)ptr_rqst_temp->packet)->offset < ((read_request*)ptr_rqst_selected->packet)->offset) && (ptr_rqst_temp->time_position < time_position_condition) ){
#if SHOWINFO > 2
	 						fprintf(stderr,YELLOW_COLOR_CHARACTERS"aiolid.c : IO_thr() selected offset = %ld (from process=%d), new offset= %ld (from process=%d)\n"END_COLOR_CHARACTERS,
										 ((read_request*)ptr_rqst_selected->packet)->offset,((read_request*)ptr_rqst_selected->packet)->pid_desc,
										 ((read_request*)ptr_rqst_temp->packet)->offset,((read_request*)ptr_rqst_temp->packet)->pid_desc);
#endif 						  
							ptr_rqst_selected=ptr_rqst_temp;
							break;
						}		
					}

#if SHOWINFO > 2
			// Parse the whole request stored in related read list.
			// TODO : Notes : Only READ_REQ cas has been implemented 
					fprintf(stderr,BLUE_COLOR_CHARACTERS"aiolid.c : IO_thr() aggregate step : There are more than one request in queue\n"END_COLOR_CHARACTERS);
					fprintf(stderr,BLUE_COLOR_CHARACTERS"aiolid.c : IO_thr() req  selected : offset = %ld, length = %d, time_position = %llu\n"END_COLOR_CHARACTERS,
									((read_request*)ptr_rqst_selected->packet)->offset,((read_request*)ptr_rqst_selected->packet)->length,ptr_rqst_selected->time_position);
				
					list_for_each_safe(ptr_position,ptr_copy_position,ptr_related_list_head){
						ptr_rqst_temp=list_entry(ptr_position,rqst_cell,list);
						fprintf(stderr,BLUE_COLOR_CHARACTERS"aiolid.c : IO_thr() list read req : offset = %ld, length=%ld, pid_desc=%d, time_position = %llu\n"END_COLOR_CHARACTERS,
									  ((read_request*)ptr_rqst_temp->packet)->offset,((read_request*)ptr_rqst_temp->packet)->length,
									  ((read_request*)ptr_rqst_temp->packet)->pid_desc,ptr_rqst_temp->time_position);
					}
#endif
	
				// Build the specific vector of requests.
					INIT_LIST_HEAD(&temp_list);
				
				// Retrieve backward and forward reference in order to start aggregate process
					if(ptr_rqst_selected->list.prev!=ptr_related_list_head)
						ptr_rqst_backward=list_entry(ptr_rqst_selected->list.prev,rqst_cell,list);

					if(ptr_rqst_selected->list.next!=ptr_related_list_head)
						ptr_rqst_forward=list_entry(ptr_rqst_selected->list.next,rqst_cell,list);
		
				// Add the selected request
					assert(ptr_rqst_selected->state==FILLED);
					list_del(&(ptr_rqst_selected->time_list));
					list_del(&(ptr_rqst_selected->list));
					ptr_rqst_selected->state=USED;
					list_add(&(ptr_rqst_selected->list),&temp_list);
							
					index=1; // Index is used to find the number of aggregated request(s)
			
				// Backward step : each request is aggregated at the beginning (recyprocally for the forward step, each request is aggregated at the end of list).
				// In this way, we keep the ascending order - stupid but easy tip ;)
				// Notes : read aggregation and write aggregation are handled by different constraints. For example, overloaded read request could be aggregate but write must'nt
					ptr_rqst_temp=ptr_rqst_selected;
#if SHOWINFO > 2
					if(ptr_rqst_backward == NULL)
						fprintf(stderr,BLUE_COLOR_CHARACTERS"aiolid.c : IO_thr() aggregate step: No backward process\n"END_COLOR_CHARACTERS);
#endif			
					while((ptr_rqst_backward != NULL) &&
						 	(((ptr_rqst_temp->type==READ_REQ)&&(isReadAggregatable(ptr_rqst_backward,ptr_rqst_temp))) ||
					 		 ((ptr_rqst_temp->type==WRITE_REQ)&&(isWriteAggregatable(ptr_rqst_backward,ptr_rqst_temp))) ) ){

						assert(ptr_rqst_backward->state==FILLED);
					
						ptr_copy_position=ptr_rqst_backward->list.prev; // As for_each_safe, store reference before delete it
						if(ptr_rqst_backward->time_position < time_position_condition){ // if the request could be execute before the next write	
							list_del(&(ptr_rqst_backward->time_list));
							list_del(&(ptr_rqst_backward->list));
			
							ptr_rqst_backward->state=USED;
							list_add(&(ptr_rqst_backward->list),&temp_list);
#if SHOWINFO > 2
							fprintf(stderr,"aiolid.c:IO_thr(): find a request to aggregate during the backward process\n");	
#endif 				  
							index++;
						}
						if(ptr_copy_position!=ptr_related_list_head){
							ptr_rqst_temp=ptr_rqst_backward;
							ptr_rqst_backward=list_entry(ptr_copy_position,rqst_cell,list);
						}
						else{
							ptr_rqst_backward=NULL;
							ptr_rqst_temp=ptr_rqst_selected;
						}	
					}

				// Forward step
#if SHOWINFO > 2
					if(ptr_rqst_forward == NULL)
						fprintf(stderr,BLUE_COLOR_CHARACTERS"aiolid.c : IO_thr() aggregate step: No forward process\n"END_COLOR_CHARACTERS);
#endif			

					while(( ptr_rqst_forward != NULL) &&
						 	( ((ptr_rqst_temp->type==READ_REQ)&&(isReadAggregatable(ptr_rqst_temp,ptr_rqst_forward))) ||
					 		  ((ptr_rqst_temp->type==WRITE_REQ)&&(isWriteAggregatable(ptr_rqst_temp,ptr_rqst_forward))) ) ){
					
						ptr_copy_position=ptr_rqst_forward->list.next; // As for_each_safe, store reference before delete it
						if(ptr_rqst_forward->time_position < time_position_condition){ // if the request could be execute before the next write	
							list_del(&(ptr_rqst_forward->time_list));
							list_del(&(ptr_rqst_forward->list));
			
							ptr_rqst_forward->state=USED;
							list_add_tail(&(ptr_rqst_forward->list),&temp_list);
#if SHOWINFO > 2
							fprintf(stderr,"aiolid.c:IO_thr(): find a request to aggregate during the forward process\n");	
#endif 				  
							index++;
						}

						if(ptr_copy_position!=ptr_related_list_head){
							ptr_rqst_temp=ptr_rqst_forward;
							ptr_rqst_forward=list_entry(ptr_copy_position,rqst_cell,list);
						}
						else	
							ptr_rqst_forward=NULL;
					}
			
				// Here, temp_list is fill in an offset ordered.	
#if SHOWINFO > 2
					fprintf(stderr,"aiolid.c:IO_thr(): temp_list is ordered and contains %d request (list contained %d request)\n",index,fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].readrqst_nb);	
#endif 				  
					if (ptr_rqst_selected->type==READ_REQ){
						fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].readrqst_nb-=index;
						if(pthread_mutex_unlock(&rqsts_mutex)!=0){
							DIE("aiolid.c: pthread_mutex_unlock Error");
						}
						if(index==1){ // No request have been aggregated so just process the selected
							process_readrqst((read_request*)ptr_rqst_selected->packet);
#if SHOWINFO > 2
							fprintf(stderr,YELLOW_COLOR_CHARACTERS"aiolid.c:IO_thr(): Only one READ_REQ request has been proccessed after aggregating process.\n"END_COLOR_CHARACTERS);	
#endif 				  
			 				fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].aggregate_readprocess=0;
						}
						else{
#if SHOWAGGREGATED
		  			 		fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].readaggregatedrqst_nb+=index;	// Just to know how many request have been aggregated for this file (to first open call untill last close call).
		  			 		if (fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].maxread_aggolmeration<index)
		  			 			fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].maxread_aggolmeration=index;
#endif						
			 				fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].aggregate_readprocess=index;
							process_readrqsts(&temp_list);
							//wait_aggregate();
						}
					}
					else if(ptr_rqst_selected->type==WRITE_REQ){
						fdesc_request[((write_request*)ptr_rqst_selected->packet)->daemonfd].writerqst_nb-=index;
						if(pthread_mutex_unlock(&rqsts_mutex)!=0){
							DIE("aiolid.c: pthread_mutex_unlock Error");
						}
						if(index==1){ // No request have been aggregated so just process the selected
							process_writerqst((write_request*)ptr_rqst_selected->packet);
			 				fdesc_request[((write_request*)ptr_rqst_selected->packet)->daemonfd].aggregate_writeprocess=0;
						}
		  				else{
#if SHOWAGGREGATED
		  			  		fdesc_request[((read_request*)ptr_rqst_selected->packet)->daemonfd].writeaggregatedrqst_nb+=index;	// Just to know how many request have been aggregated for this file (to first open call untill last close call).	
#endif 			  
							process_writerqsts(&temp_list);
						}
					}
					
					// Add proccessed request(s) in the free_list	
					if(pthread_mutex_lock(&rqsts_mutex)!=0){
						DIE("aiolid.c: pthread_mutex_lock Error");
					}
					list_for_each_safe(ptr_position,ptr_copy_position,&temp_list){
						ptr_rqst_selected=list_entry(ptr_position,rqst_cell,list);
						ptr_rqst_selected->state=FREE;
						list_del(&ptr_rqst_selected->list); // delete from temp list : should be useless since we use a safe iteration (for_each_safe) but I have to confirm, need time...
						list_add(&(ptr_rqst_selected->list),&free_list);	// And add into free list
					}
#if SHOWINFO > 1
					fprintf(stderr,"aiolid.c:IO_thr(): %d requests (with type=%ld) have been proccessed. %d new free cell(s) have been added to the free_list\n",index,ptr_rqst_selected->type,index);	
#endif 				  
				}	
#if SHOWINFO > 1
			fprintf(stderr,"\n");
#endif 				  
			// Wake up receiver thread
			if(pthread_cond_signal(&rqst_cond_free)!=0){
				DIE("aiolid.c: pthread_cond_signal Error");
			}
			if(pthread_mutex_unlock(&rqsts_mutex)!=0){
				DIE("aiolid.c: pthread_mutex_lock Error");
			}
		} // end of READ_REQ / WRITE_REQ part

	}// end of while(1)
}

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
//
// MAIN AND ARGS ANALYSE
//
// // // // // // // // // // // // // // // // // // // // // // // // // // // // // 
// Global Variables
static int nioservers=1; 	// Corresponding  to the number of thread used for make real IO.
char* logfile=NULL;
																		
//static int foreground = 0;

static void usage(FILE* const f) {
		fprintf(f,
"Usage: " PROGNAME " [options] datapath\n"
"\n"
"Launches an aioli daemon (aiolid) chdir()'ed in datapath\n"
"\n"
" -h, --help          this page\n"
" -F, --foreground    don't go background           (default: go background)\n"
" -m x, --msg_queuekey x      use ipc message queue defined by key :x 	(default: 1234)\n"
" -n x, --nservers x  spawns x I/O servers                      (default: 1)\n"
" -d, --debug         verbose debug mode                      (default: off)\n"
" -l file, --logfile file	redirect stdout/stderr to file	(default: no redirection)\n"
"\n"
		);
		exit( (f == stderr) ? 1 : 0);
}


void parse_command_line(int argc, char *argv[]) {
	static const char * shortopts = "hFm:n:dl:";
	static const struct option longopts[] = {
		{ "help",	0,			0, 'h' },
		{ "foreground",	0,			0, 'F' },
		{ "msg_queuekey",	required_argument,	0, 'm' },
		{ "nservers",	required_argument,	0, 'n' },
		{ "debug",	0,			0, 'd' },
		{ "logfile",	required_argument,	0, 'l' },
		{ "debug",	0,			0, 'd' },
		{ NULL,		0,			0, 0 },
	};
	char c;

//	Dprintf(D_IOD, "About to parse command line");

	opterr = 0;
	while((c = getopt_long(argc, argv, shortopts, longopts, NULL)) != EOF) {
		switch(c) {
			case 'h':
				usage(stdout);
				break;
			case 'F':
	//			foreground = 1;
				break;
			case 'm':
				msg_queuekey = atoi(optarg);
				if (msg_queuekey <= 0) {
					fprintf(stderr, PROGNAME ": Could'nt use %s for the message queue key\n", optarg);
					exit(1);
				}
				break;
			case 'n':
				nioservers = atoi(optarg);
				if (nioservers <= 0) {
					fprintf(stderr, PROGNAME ": there must be at least 1 server: %s\n", optarg);
					exit(1);
				}
				break;
			case 'd':
				//enable_logging("iod");
				break;
			case 'l':
				logfile = malloc((sizeof(char)+1)*strlen(optarg));
				strcpy(logfile,optarg);
				break;

			default:
				fprintf(stderr, PROGNAME ": unknown option '%c'\n", c);
				usage(stderr);
		}
	}

/*	if (argc - optind == 1) {
		datapath = argv[optind];
	}
  	else {
		fprintf(stderr, PROGNAME ": incorrect number of arguments\n");
		exit(1);
	}*/
}

int main(int argc, char *argv[]) {
	
	int index; 
	pthread_t thr;
	
	char aiolidstat[100]; //Store PID/ GUID/ UID .... 
	int aiolifd;
	struct stat tempbuf; // Required to test if aiolid is already running
	
	sprintf(aiolidstat,"main(pid=%d): running as uid=%d euid=%d gid=%d egid=%d\n", getpid(), getuid(), geteuid(), getgid(), getegid());
	
	if (stat("/tmp/aiolid.pid",&tempbuf)>=0){
		fprintf(stdout,"aiolid.c: aiolid seems to be already run, if not please killall aiolid processus, remove /tmp/aioli.pid file and f it still persits clean related IPC objects (see with ipcs)\n");
		exit(0);
	}
	else{
		aiolifd=creat("/tmp/aiolid.pid",O_WRONLY|0666);
		if(aiolifd==-1){
			fprintf(stdout,"aiolid.c: Can't create /tmp/aiolid.pid file, so die...\n");
			perror("");
			exit(0);
		}
		 else
			write(aiolifd,aiolidstat,sizeof(char)*strlen(aiolidstat));
		close(aiolifd);
	}
	
	parse_command_line(argc, argv);
	
	if (logfile!=NULL){
	    aiolifd=open(logfile,O_CREAT|O_RDWR|O_TRUNC,0666) ;
		if(aiolifd<0){
			fprintf(stdout,"aiolid.c: Can't create %s file\n",logfile);
			perror("");
			system ("rm /tmp/aiolid.pid");
			exit(0);
		}
		 else{ // redirect stdout and stderr
			close(1); //stdout
			close(2); //stderr
			dup2(aiolifd,1);
			dup2(aiolifd,2);
			close(aiolifd);
			free(logfile);
		}
	}
	
	fprintf(stderr, "Compilation : " __DATE__ " at " __TIME__ " with " __VERSION__ "\n");
	fprintf(stderr,"%s\n",aiolidstat);
#if HAS_PRIO 
	fprintf(stderr, "receiver_thr() priority: %d, IO_thr() priority: %d \n",PRIO_RECEIVER_THR,PRIO_IO_THR);
#endif
	
//----------------------------------------------------------

	// Open IPC Message queue to communicate with client process
	msgqid=init_request_msgqueue(msg_queuekey);
	if (msgqid<0)
	  DIE("Can't find message queue so DIE :((");
			  
	// init process and file management parts
	init_tabs();

	//init list(s) : Remember this call have to be executed before thread creation ! 
	init_lists();
	
	init_shm_objs();

//----------------------------------------------------------
/*
// Create thread wich receives I/O request from client(s)
	if (pthread_create(&thr, NULL, receiver_thr, NULL)) {
//		Dprintf(L_FATAL, "aiolid.c:main(): cannot pthread_create(receiver_thr)");
//		exit(1);
		DIE("aiolid.c:main(): cannot pthread_create(receiver_thr)");
	}
	if (pthread_detach(thr)) {
//		Dprintf(L_FATAL, "aiolid.cmain(): cannot pthread_detach(receiver_thr)");
//		exit(1);
		DIE("aiolid.c:main(): cannot pthread_detach(receiver_thr)");
	}
*/
//----------------------------------------------------------

// Create thread(s) which make(s) the related I/O/
// The basic idea consist of create as many thread as the number of I/O server. The purpose is to have just one thread making one I/O on a server at a given time.
// TODO Use a mutex by I/O thread.
	for(index = 0; index < nioservers; index++) {
		if (pthread_create(&thr, NULL, IO_thr, NULL)) {
//			Dprintf(L_FATAL, "aiolid.c:main(): pthread_create(IO_thr) failed");
//			exit(1);
			DIE("aiolid.c:main(): pthread_create(IO_thr) failed");
		}
		if (pthread_detach(thr)) {
//			Dprintf(L_FATAL, "aiolid.c:main(): cannot pthread_detach(IO_thr)");
//			exit(1);
		}
	}
//	pause();
	receiver_thr(NULL);
	return 0;
}
