/*
 * File : 
 * 	daemon/aiodid.h
 * 	
 * 	aIOLi is an efficient I/O library for parallel access to remote storages in SMP clusters. 
 * 	Thanks to the kernel features, our framework provides parallel I/O algorithms without 
 * 	inter-processes synchronisation mechanisms as well as a simple interface based on 
 * 	the classic UNIX system calls (creat/open/read/write/close).
 *
 * 	This file is included in the first version of our prototype aIOLi.
 * 	To know which version is currently stable please visit http://aioli.imag.fr
 * 	
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author : 
 * 	Adrien Lebre, adrien.lebre@imag.fr
 *
 * Licence : 
 * 	GPLv2	
 * 	This program is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU General Public License
 * 	as published by the Free Software Foundation; either version 2
 * 	of the License, or (at your option) any later version.
 * 	
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program; if not, write to the Free Software
 * 	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Contributors :
 * 	BULL Open Source Division - INRIA France
 * 	
 * Created :
 * 	April 2004, the 1st
 * 
 * Last Modification(s) : 
 * 	April 2005
 * 	
 * Compilation :
 * 	 see Makefile
 * 	
 * Comments : 
 * 	Defines different structures used in aIOLi daemon.
 * 	Included aioli-shared.h
 *
 * Changelog :
 * 	look into the CVS log !
 *
*/

#ifndef __AIOLID_H__
#define __AIOLID_H__ 1

#include "../include/aioli-shared.h"

#define SHOWINFO 1
#define SHOWAGGREGATED 1
#if SHOWINFO && (! SHOWAGGREGATED)
	#define SHOWAGGREGATED 1
#endif


#define MAX_AGGREGATED_PROCESSES 20					// The maximum number of process that could used at the same time aioli daemon.
#define MAX_FILE_PERPROCESS MAX_FILEOPEN			// TODO find a way to improve (dynamically or not) the maximum number
																//	of files that could be open for aioli daemon. (Please see include <bits/stdio_lim.h> and #define FOPEN_MAX)
																//	Please, see aioli-shared.h MAX_FILEOPEN

// **** ******************************** **** 
// ****  		 Thread part              ****
// **** ******************************** **** 

#define HAS_ROOT 1
#define HAS_PRIO 1	 

#if HAS_PRIO  && (! HAS_ROOT)
	#error "priority tweaking requires root permissions"
#endif

#if HAS_ROOT && HAS_PRIO
// What should be the NICE adjustement for the threads ?
	#define PRIO_RECEIVER_THR 	-20
	#define PRIO_IO_THR 19 
//	#define PRIO_IO_HIGH 12 
//	#define PRIO_IO_LOW 14 
#endif
															
																
#include <sys/types.h>

// **** ******************************** **** 
// **** 	  	   Process management        ****
// **** ******************************** **** 
// This struct stores required informations for each client wich is using aioli service
typedef struct process_values{	
	pid_t pid;										// Calling process pid, pid >=0 means that this struct is correctly filled in !
	int msgq_replyid;								// Replies's message queue id (from aioli daemon to calling process),
														//	the message queu key have to be post by an "init_request" by the calling
														//	 process at the beginning.	
														
	int nb_file;									// Number of file open. Use this variable to handle the process_iob table.
	process_iob iob[MAX_FILE_PERPROCESS]; 	// Extended the "iob" concepts. Please see process_iob structure in aioli-shared.h. 
	int last_ack_request;						// Stores the last acknowledged request.
} process_values;


// **** ******************************** **** 
// ****    File and request management   **** 
// **** ******************************** **** 


// For each file descriptor, we have two linked list : one for the read requests and one for the writes.
// To update or make some changes in a list, you have to hold the global mutex 
#include "list.h"
typedef struct list_head rqst_list;			// Please see "./list.h"

typedef enum {FREE, RECV, FILLED, USED} staterqst_t;

// The cell for the lists.
typedef struct rqst_cell{
	rqst_list time_list;					// The references from/to the list sort by ascending timestamp.		
	u_int64_t time_position ;				// The position in the time_list (for each new request time_position is incremented)	
	rqst_list list;						//	The references from/to the assigned linked list (INIT list, OPEN list, CLOSE list ...).  			
	long type; 								// The type of the stored request (INIT_REQ,OPEN_REQ, ...)	
	char packet[MAX_PACKETSZ];			// Store the related request.
	staterqst_t state; 					// We need this king of value in order to know the state of the cell. 
} rqst_cell;

#define MAX_RQSTNB 255						// Maximum number of request
rqst_cell rqst_cells[MAX_RQSTNB]; 		// This tab provides a way to bypass malloc and free calls ;) to store dynamically each request
#include <pthread.h>
//#include "../GLIBC-2.3.4/glibc-2.3.4/include/pthread.h"
pthread_mutex_t rqsts_mutex;     		// Before to access to different list, you have to hold the rqsts_mutex
static pthread_cond_t  rqst_cond_free; 
static pthread_cond_t  rqst_cond_filled;
// Notes : We don't need a global predicate as freerqst_number and filledrqst_number. Actually we used list_empty(free_list) and list_empty(time_order_list) wich are similar

rqst_list free_list; 		// This list stores free cells
rqst_list time_order_list; 	// This list stores the whole requests in a time order.

												
typedef struct file_request{
		  //	int file_desc;			// The descriptor returned by creat or open in aioli daemon.
											//	Actually we're using a tab of file_request (as fdesc_info, see below), the index corresponds to the file descriptor !
	rqst_list readrqst_list;		// Each read request to a defined file are put in this linked list.
	int readrqst_nb;					// Number of request contained in readrqst_list
	int aggregate_readprocess;		// Stand for the number of request wich have been aggregated during last access to readrqst_list : 0 means any aggregate process has started,
  											//	1, could'nt exist ;), 2 means two read requests, 3 for 3 requests and so on, .....
#if SHOWAGGREGATED
	unsigned long readaggregatedrqst_nb;		// Just to know how many request have been aggregated for this file (to first open call untill last close call).	
	int maxread_aggolmeration;							// Just to know the biggest aglomeration
#endif			  
	rqst_list writerqst_list;		// Each write request to a defined file are put in this linked list.
	int writerqst_nb;					// Number of request contained in writerqst_list
	int aggregate_writeprocess;	// See aggregate_readprocess above.
#if SHOWAGGREGATED
	unsigned long writeaggregatedrqst_nb;		// Just to know how many request have been aggregated for this file (to first open call untill last close call).	
#endif			  
} file_request;


// Required informations for each opened file : 
typedef struct file_desc_info{			
  	
	//	int file_desc;								// The descriptor returned by creat or open in aioli daemon.
														//	Actually we're using a tab of fdesc_info, the index represents the file_desc !
	char pathname[MAX_FILENAMESZ];			// absolute filename : /.....
	// TODO Add a hash representation of pathname, it should be more efficient than compare two string. (for example, take in accoutn 32 bits of sha1 from openssl/sha.h)
	// Please see open_file_desc() and add_file() functions in aioli.c
	
	int nb_owners;									// Number of process which are working on this file.
	
	int owners[MAX_AGGREGATED_PROCESSES];	// List of these processes.

	off_t global_position;						// Current position for the file descriptor used by aioli daemon.

	int flags;										// Flags argument for creat or open calls. 
	mode_t mode;									// as flags, mode argument.

	pthread_mutex_t open_close_mutex;	 	// You have to hold this mutex for all open/close requests on file fdesc_tab[i] 
		
} fdesc_info;




// **** ******************************** **** 
// ****       Prototypes part            ****
// **** ******************************** **** 

void clean_ipc(void);

int open_file_desc(int desc_pid, const char* pathname, int flags, int mode, int userfd);
int close_file_desc(int desc_pid, int userfd, int daemonfd);


int init_reply_msgqueue(key_t key);
int post_genericreply(int msgqid,int rqstid, long type, int result, int error);
int post_reply(int msg_queueid, long type,void *packet, size_t packet_size);

void init_tabs (void);
int add_process (pid_t pid, key_t msg_replykey,const char *shm_name);
int remove_process (int pid_desc,pid_t pid);

// Add file informations inside fdesc_tab variable
int add_file(int pid_desc, int userfd, const char* pathname, int daemonfd);
int remove_file(int pid_desc, int userfd, int daemonfd);


#endif

