/*
 * File : 
 * 	include/aioli-shared.h
 * 	
 * 	aIOLi is an efficient I/O library for parallel access to remote storages in SMP clusters. 
 * 	Thanks to the kernel features, our framework provides parallel I/O algorithms without 
 * 	inter-processes synchronisation mechanisms as well as a simple interface based on 
 * 	the classic UNIX system calls (creat/open/read/write/close).
 *
 * 	This file is included in the first version of our prototype aIOLi.
 * 	To know which version is currently stable please visit http://aioli.imag.fr
 * 	
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author : 
 * 	Adrien Lebre, adrien.lebre@imag.fr
 *
 * Licence : 
 * 	GPLv2	
 * 	This program is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU General Public License
 * 	as published by the Free Software Foundation; either version 2
 * 	of the License, or (at your option) any later version.
 * 	
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program; if not, write to the Free Software
 * 	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Contributors :
 * 	BULL Open Source Division - INRIA France
 * 	
 * Created :
 * 	April 2004, the 1st
 *
 * Last Modification(s) : 
 * 	April 2005
 *
 * Compilation :
 * 	 see Makefile
 * 	
 * Comments : 
 * 	Defines different request structures exchange from aioli client part to aiolid
 *
 * Todo : 
 *  	grep TODO in file ;) ..... :D 
 *  	create new errno value (check error string, just below).
 *  	
 * Changelog :
 * 	look into the CVS log !
 *
*/

#ifndef __AIOLI_SHARED_H__
#define __AIOLI_SHARED_H__ 1


#include <sys/types.h>
#include <errno.h>


// **** ************************** **** 
// ****			Error strings 		  ****		 
// **** ************************** **** 
#define ENPROCESS -123						// TODO Have to define ENPROCESS error which means the maximum
										 			// Number of process has been reached (similar to ENFILE ;), man 2 open).											
#define EN_FILEPERPROCESS -124			// Number of file per process has been reached.
#define EN_AGGREGATEDPROCESSES -125		// Number of aggregated processes has been reached.
#define EBAD_PIDDESC -126					// Bad pid descriptor.

#define GREEN_COLOR_CHARACTERS "\e[1;42m"
#define YELLOW_COLOR_CHARACTERS "\e[1;43m"
#define RED_COLOR_CHARACTERS "\e[1;31m"
#define BLUE_COLOR_CHARACTERS "\e[1;44m"
#define END_COLOR_CHARACTERS "\e[m"

// **** ************************** **** 
// ****	IPC Message Queue part 	  ****	 
// **** ************************** **** 
#define MESSAGE_QUEUE_KEY 1234		//	The global key to indentify the global IPC Message queue.
#define MAX_PACKETSZ 8192 				// The maximum size of a request send from a client to the aIOLi by the IPC Message queue. 
												// Notes : MAX_PACKETSZ shoudl be replace by MSGMAX define in <sys/msg.h>

// **** ************************** **** 
// **** 	 	Miscellaneous part 	  ****	 
// **** ************************** **** 
#define MAX_FILENAMESZ 4096			// Please see include <bit/stdio_lim.h> and #define FILE_NAMEMAX
#define MAX_FILEOPEN 16					// TODO find a way to improve (dynamically or not) the maximum number
												//	of files that could be open for aIOLi daemon. (Please see include <bits/stdio_lim.h> and #define FOPEN_MAX)
												//	Please, see aiolid.h MAX_FILE_PERPROCESS
												//	

// **** ************************** **** 
// **** 		Request part 			  ****
// **** ************************** **** 

// Request Types : 
// Warning : the order is important, the request have been to be sort in an priprity ascending way.
// A part of the aIOLi is based on this point. Please see aiolid.h and 'lists' tab variable.
#define INIT_REQ 0
#define OPEN_REQ	1
#define CLOSE_REQ	2
#define FINAL_REQ 3
#define READ_REQ 4
#define WRITE_REQ	5

#define NB_REQ_TYPE 6

//TODO to add reliability to our model, requestid should be 

// Request Structure
// Required for IPC message queue.
typedef struct msg_request {
	long type;
	char packet[MAX_PACKETSZ];
} msg_request;

typedef struct init_request{
	int requestid;
	pid_t pid;
	key_t msgq_replykey;
	char shm_name[10];
} init_request;

typedef struct finalize_request{
	int requestid;
	pid_t pid; 				// for "finalize", pid and pid_desc are required to valide process authentification ;). 
	int pid_desc;
} finalize_request;

typedef struct open_request{
	int requestid;
	int pid_desc;

	int userfd;
	char pathname[MAX_FILENAMESZ];
	int flags;
	mode_t mode;
} open_request;

typedef struct close_request{
	int requestid;
	int pid_desc;
	
	int userfd;
	int daemonfd;
} close_request;

typedef struct read_request{
	int requestid;
	int pid_desc;
	
	int userfd;
	off_t offset;
	size_t length;
	
	int daemonfd;
	//TODO shared memory segment id
	//....
} read_request;

typedef struct write_request{
	int requestid;
	int pid_desc;
	
	off_t offset;
	size_t length;
	int daemonfd;
	
	
} write_request;


// **** ************************** **** 
// ****			Reply part			  ****	 
// **** ************************** **** 

// Reply Types : 
#define INIT_REPLY	INIT_REQ+10
#define OPEN_REPLY	OPEN_REQ+10
#define CLOSE_REPLY	CLOSE_REQ+10
#define FINAL_REPLY	FINAL_REQ+10
#define READ_REPLY 	READ_REQ+10
#define WRITE_REPLY	WRITE_REQ+10

typedef msg_request msg_reply;
typedef struct generic_reply{				// At the first post message queue request, a specific
		 											// message queue is created and used to post
													// the reply delivered by aIOLi daemon.
	int requestid;								// The request identificator.
	int result;									// Reply from aiolid.
	int remote_errno;							// aIOLi errno value in case of.
} generic_reply ; 

typedef struct read_reply{				
		 											
	int requestid;								// The request identificator.
	int result;									// Reply from aIOLi (usually, it saves the number of read bytes).
	int remote_errno;							// aIOLi errno value in case of.

	int shm_index;								// The index to access to the right shm_obj. 
	off_t seek_start;							//	The required move to directly access to related data in the shm_obj.
} read_reply ; 

#define MAX_SHMNB 16 						// The max number of reply_shm_obj.
//#define MAX_COPYBUFFERSZ 268435456	// 256 Mbytes, the size of memory mapped segment, find ftruncate in init_shm_buffer() function
//#define MAX_COPYBUFFERSZ 2097152		// 2 Mbytes, the size of memory mapped segment, find ftruncate in init_shm_buffer() function
//#define MAX_COPYBUFFERSZ 8388608		// 8 Mbytes, the size of memory mapped segment, find ftruncate in init_shm_buffer() function
#define MAX_COPYBUFFERSZ 33554432 		// 32 Mbytes, the size of memory mapped segment, find ftruncate in init_shm_buffer() function
//#define MAX_COPYBUFFERSZ 50331648 		// 48 Mbytes, the size of memory mapped segment, find ftruncate in init_shm_buffer() function
													
													// Notes :
													// By default, Linux allows 4096 shmen id (more /proc/sys/kernel/shmmni) with a max size per segment of 32Mo (more /proc/sys/kernel/shmmax)
													// The max size per segment is not enough because we want to get only one global segment. Thus, we don't use shmget function but instead shm_open 
													// and ftruncate in order to allow a memory file object wit the wished size.
#define GLOBAL_SHM_NAME "aioli"


#include "shmmutex.h"
#define FIRST_MUTEX_KEY 4321				//	The first key used to create required semaphores to access to different reply_shm_obj.
int shm_mutex[MAX_SHMNB];

typedef struct reply_shm_obj{
	int link;									// Number of remaining processes that should access to the shm_obj.
													// 'link==0' means that shm_obj is free and could be reused.
	char bytes[MAX_COPYBUFFERSZ];
} reply_shm_obj;

char *shm_name=GLOBAL_SHM_NAME;
void* shm_buffer;
int shm_desc=-1;

//reply_shm_obj shm_objs[MAX_SHMNB];
reply_shm_obj *shm_objs; 					// Actually shm_objs[MAX_SHMNB] memory corresponds to the memory space allocate by shm_open, ftruncate and mmap calls
													// (please see init_shm_buffer() function).
// **** ************************** **** 
// ****	Process management part   ****	 
// **** ************************** **** 

typedef struct process_iob{ //TODO find a way to redirect fstdout fstdin et fstderr...
	
//	int processfd;								// file descriptor in calling process 
// Actually, useless since we use the process_desc as an index in both process (user process and aIOLi daemon).
 
	off_t process_position; 					// offset position for calling process
	int daemonfd;									// file descriptor in aIOLi daemon 

	// Actually, doesn't seem to be usefull at this time, maybe we will use to implement shared pointer concept but today :p ...
	off_t daemon_position;			   		// offset position for aIOLi daemon (logically is equals to global_position from fdesc_info structure, see aiolid.h)
														// Actually, process_position and daemon_position should be the same value
														// but as in MPI IO specifications we plan to implement something like	
														// shared and individual file pointer.
} process_iob;

#endif

