/* File : 
 * 	include/shmmutex.h
 * 	
 * 	aIOLi is an efficient I/O library for parallel access to remote storages in SMP clusters. 
 * 	Thanks to the kernel features, our framework provides parallel I/O algorithms without 
 * 	inter-processes synchronisation mechanisms as well as a simple interface based on 
 * 	the classic UNIX system calls (creat/open/read/write/close).
 *
 * 	This file is included in the first version of our prototype aIOLi.
 * 	To know which version is currently stable please visit http://aioli.imag.fr
 * 	
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author : 
 * 	Adrien Lebre, adrien.lebre@imag.fr
 *
 * Licence : 
 * 	GPLv2	
 * 	This program is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU General Public License
 * 	as published by the Free Software Foundation; either version 2
 * 	of the License, or (at your option) any later version.
 * 	
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program; if not, write to the Free Software
 * 	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Contributors :
 * 	BULL Open Source Division - INRIA France
 * 	
 * Created :
 * 	August 2004, the 10st
 *
 * Last Modification(s) : 
 * 	April 2005
 * 	
 * Compilation :
 * 	 see Makefile
 * 	
 * Comments : 
 * 	Implement mutex mechanism based on IPC semaphores
 * 	in order to exchange data thanks to shared memory segments between user processes and aIOLi daemon.
 * 	
 * 	Actually, to access to a specific shared memory segment, you have to hold the related mutex. Please see daemon/aiolid.h and daemon/aiolid.c
 * 	aIOLi uses this shared segment to put each related data from aiolid to client part.
 * 	The post_readrqst() have just to copy from shared memory segment to the real buffer provided by read_aioli call.
 * 	Note this approach is simple but not thread safe. Indeed, if two thread make a read, the shared memory segment is n't protected.
 *
 * Todo : 
 * 	grep TODO in file ;) ..... :D 
 * 	handle thread_safe constraints.
 *
 * Changelog :
 * 	look into the CVS log !
 *
*/

#ifndef _SHM_MUTEX_AIOLI_H
#define _SHM_MUTEX_AIOLI_H

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <stdio.h>
#include <errno.h>


// init a semaphore and set semval to 1.
static inline int mutex_init(key_t key){
	int semaid=-1;
	union semun {
		int val;
		struct semid_ds *buf;
		ushort * array;
	} argument;
   argument.val = 1;
	
	semaid=semget(key,1,IPC_CREAT|0666);
	if(semaid==-1){
		fprintf(stderr,"shmmutex.h: mutex_init(): Can't initialize a semaphore with key %d : semget() Error\n",key);
		perror("");
	}		

	if(semctl(semaid,0,SETVAL,argument)<0){
		fprintf(stderr,"shmmutex.h: mutex_init(): Can't set 1 to semval of semaiid=%d : semget() Error\n",semaid);
		perror("");
	}
#if SHOWINFO > 2 
	fprintf(stderr,"shmmutex.h : mutex_init(): a new semaphore (id=%d, key=%d) has been initialized\n",semaid,key);
#endif
	return semaid;
}

// remove a semaphore create by mutex_init.
static inline int mutex_remove(int semaid){
	int result=0;
	result=semctl(semaid,0,IPC_RMID);
	if(result<0){
		fprintf(stderr,"shmmutex.h: mutex_remove(): Can't remove semaid=%d (result=%d): semctl() Error errno=%d\n",semaid,result,errno);
		perror("");
		return result;
	}
#if SHOWINFO > 3 
	fprintf(stderr,"shmmutex.h : mutex_remove(): semaphore (id=%d) has been removed\n",semaid);
#endif
	return result;
}

// retrieve a semaphore id from a key.
static inline int mutex_get(key_t key){
	int semaid=-1;
	semaid=semget(key,1,0666);
	if(semaid==-1){
		fprintf(stderr,"shmmutex.h: mutex_get(): Can't get a semaphore with key %d : semget() Error\n",key);
		perror("");
	}		
#if SHOWINFO > 2 
	fprintf(stderr,"shmmutex.h : mutex_get(): get semaphore (id=%d, key=%d)\n",semaid,key);
#endif
	return semaid;
}

static inline int mutex_lock(int semaid){
	int retval=-1;	
	struct sembuf operations[1];
	operations[0].sem_num = 0;
	operations[0].sem_op = -1;
	operations[0].sem_flg = SEM_UNDO;
	retval = semop(semaid, operations, 1);
#if SHOWINFO
	if(retval != 0){			  
		fprintf(stderr,"shmmutex.h:mutex_lock(): can't lock semaid=%d :semop Error\n",semaid);
		perror("");
	}
#endif
	return retval;
}

static inline int mutex_unlock(int semaid){
	int retval=-1;	
	struct sembuf operations[1];
	operations[0].sem_num = 0;
	operations[0].sem_op = 1;
	operations[0].sem_flg = SEM_UNDO;
	retval = semop(semaid, operations, 1);
#if SHOWINFO
	if(retval != 0){			  
		fprintf(stderr,"shmmutex.h:mutex_unlock(): can't lock semaid=%d :semop Error\n",semaid);
		perror("");
	}
#endif
	return retval;
}

#endif
